# Copyright 2018 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import

import logging
from http import client as http_client

from restalchemy.api import middlewares
from restalchemy.common import exceptions as comm_exc
from restalchemy.storage import exceptions as ra_exceptions

UNKNOWN_ERROR_CODE = 10000001


LOG = logging.getLogger(__name__)


def exception2dict(exception):
    code = (
        exception.get_code()
        if hasattr(exception, "get_code")
        else (exception.code if hasattr(exception, "code") else UNKNOWN_ERROR_CODE)
    )
    message = (
        exception.msg
        if hasattr(exception, "msg")
        else (exception.message if hasattr(exception, "message") else str(exception))
    )
    return {
        "type": exception.__class__.__name__,
        "code": code,
        "message": message,
    }


class ErrorsHandlerMiddleware(middlewares.Middleware):
    not_found_exc = (ra_exceptions.RecordNotFound,)
    conflict_exc = (ra_exceptions.ConflictRecords,)
    common_exc = (comm_exc.RestAlchemyException,)
    valid_exc = (comm_exc.ValidationErrorException,)

    def _construct_error_response(self, req, e):
        if isinstance(e, self.not_found_exc):
            return req.ResponseClass(
                status=http_client.NOT_FOUND, json=exception2dict(e)
            )
        elif isinstance(e, self.conflict_exc):
            return req.ResponseClass(
                status=http_client.CONFLICT, json=exception2dict(e)
            )
        elif isinstance(e, self.valid_exc):
            return req.ResponseClass(
                status=http_client.BAD_REQUEST, json=exception2dict(e)
            )
        elif isinstance(e, self.common_exc):
            return req.ResponseClass(status=e.code, json=exception2dict(e))
        else:
            return req.ResponseClass(
                status=http_client.INTERNAL_SERVER_ERROR,
                json=exception2dict(e),
            )

    def process_request(self, req):
        try:
            return req.get_response(self.application)
        except Exception as e:
            LOG.exception(
                "Unknown error has occurred on url: %s %s", req.method, req.url
            )
            return self._construct_error_response(req, e)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
#    Copyright 2023 v.burygin
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from wsgiref.simple_server import make_server

from restalchemy.api import actions
from restalchemy.api import applications
from restalchemy.api import controllers
from restalchemy.api import middlewares
from restalchemy.api import resources
from restalchemy.api import routes
from restalchemy.api import constants

from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types

from restalchemy.openapi import engines as openapi_engines
from restalchemy.openapi import structures as openapi_structures
from restalchemy.openapi import utils as oa_utils
from restalchemy.openapi import constants as oa_c

HOST = "0.0.0.0"
PORT = 8000


# There's an example of custom openapi spec, see File* models, controllers and
#   routes. If you don't need that - just skip them.


class FileModel(models.ModelWithUUID):
    name = properties.property(types.String(), required=True)


# Example controller with custom openapi specs
class FilesController(controllers.Controller):
    """Controller for /v1/files/[<id>] endpoint"""

    __resource__ = resources.ResourceByRAModel(
        model_class=FileModel,
        process_filters=True,
        convert_underscore=False,
    )

    def create(self, **kwargs):
        # Implement file upload logic here
        pass

    # Example of custom openapi spec for file upload/download,
    #   see restalchemy.openapi.constants for other request-response templates
    create.openapi_schema = oa_utils.Schema(
        summary="Upload file",
        parameters=(),
        responses=oa_c.build_openapi_create_response(
            "%s_Create" % __resource__.get_model().__name__
        ),
        request_body=oa_c.build_openapi_req_body_multipart(
            description="Upload file to docs set",
            properties={"file": {"format": "binary", "type": "string"}},
        ),
    )

    # There's decorator too, but openapi_schema attribute usage
    #   is recommended, see below.
    @oa_utils.extend_schema(
        summary="Download file",
        parameters=(),
        responses=oa_c.build_openapi_response_octet_stream(),
    )
    @actions.get
    def download(self, resource, **kwargs):
        # Some function which returns file data
        data = resource.download_file()
        headers = {
            "Content-Type": constants.CONTENT_TYPE_OCTET_STREAM,
            "Content-Disposition": 'attachment; filename="%s"' % resource.name,
        }

        return data, 200, headers

    download.openapi_schema = oa_utils.Schema(
        summary="Download file",
        parameters=(),
        responses=oa_c.build_openapi_response_octet_stream(),
    )


class FileDownloadAction(routes.Action):
    """Handler for /v1/files/<id>/actions/download endpoint"""

    __controller__ = FilesController


class FilesRoute(routes.Route):
    """Handler for /v1/files/<id> endpoint"""

    __controller__ = FilesController
    __allow_methods__ = [
        routes.CREATE,
        routes.DELETE,
        routes.FILTER,
        routes.GET,
        routes.UPDATE,
    ]

    download = routes.action(FileDownloadAction, invoke=True)


class RootController(controllers.RoutesListController):
    """Controller for / endpoint"""

    __TARGET_PATH__ = "/"


class ApiEndpointController(controllers.RoutesListController):
    """Controller for /v1/ endpoint"""

    __TARGET_PATH__ = "/v1/"


class ApiEndpointRoute(routes.Route):
    """Handler for /v1/ endpoint"""

    __controller__ = ApiEndpointController
    __allow_methods__ = [
        routes.FILTER,
        routes.GET,
    ]

    specifications = routes.action(routes.OpenApiSpecificationRoute)
    files = routes.route(FilesRoute)


class UserApiApp(routes.RootRoute):
    __controller__ = controllers.RootController
    __allow_methods__ = [routes.FILTER]


# Route to /v1/ endpoint.
setattr(
    UserApiApp,
    "v1",
    routes.route(ApiEndpointRoute),
)


def get_openapi_engine():
    openapi_engine = openapi_engines.OpenApiEngine(
        info=openapi_structures.OpenApiInfo(),
        paths=openapi_structures.OpenApiPaths(),
        components=openapi_structures.OpenApiComponents(),
    )
    return openapi_engine


def get_user_api_application():
    return UserApiApp


def build_wsgi_application():
    return middlewares.attach_middlewares(
        applications.OpenApiApplication(
            route_class=get_user_api_application(),
            openapi_engine=get_openapi_engine(),
        ),
        [],
    )


def main():
    """

    After start you can try curl http://127.0.0.1:8000/v1/specifications/3.0.3

    """
    server = make_server(HOST, PORT, build_wsgi_application())

    try:
        print("Serve forever on %s:%s" % (HOST, PORT))
        server.serve_forever()
    except KeyboardInterrupt:
        print("Bye")


if __name__ == "__main__":
    main()



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import  # noqa

from functools import wraps

from mysql.connector import errors

from restalchemy.storage.sql.dialect import base
from restalchemy.storage.sql.dialect import exceptions as exc

SAVEPOINT_EXP_MAP = {
    "savepoint": "SAVEPOINT {name};",
    "rollback": "ROLLBACK TO {name};",
    "release": "RELEASE SAVEPOINT {name};",
}


def handle_database_errors(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """
        A decorator to handle database errors.

        This decorator catches database errors from the
        ``mysql-connector-python`` library and raises them as
        :class:`.DeadLock` or :class:`.Conflict` exceptions when appropriate.

        :param func: The function to be wrapped.
        :type func: callable
        :return: A function that wraps `func` and handles database errors.
        :rtype: callable
        """
        try:
            return func(self, *args, **kwargs)
        except errors.DatabaseError as e:
            if e.errno == 1213:
                raise exc.DeadLock(code=e.sqlstate, message=e.msg)
            elif e.errno == 1062:
                raise exc.Conflict(code=e.sqlstate, message=e.msg)
            raise

    return wrapper


class MySQLInsert(base.BaseInsertCommand):
    @handle_database_errors
    def execute(self):
        """
        Executes the MySQL command.

        This method utilizes the base class's execute method to
        perform the command execution. It is decorated with
        `handle_database_errors` to handle any MySQL database
        errors that may occur during execution, such as deadlocks
        or conflicts, and raise appropriate exceptions.

        :return: The result of the command execution.
        """

        return super().execute()


class MySQLUpdate(base.BaseUpdateCommand):
    @handle_database_errors
    def execute(self):
        """
        Executes the MySQL command.

        This method utilizes the base class's execute method to
        perform the command execution. It is decorated with
        `handle_database_errors` to handle any MySQL database
        errors that may occur during execution, such as deadlocks
        or conflicts, and raise appropriate exceptions.

        :return: The result of the command execution.
        """
        return super().execute()


class MySQLDelete(base.BaseDeleteCommand):
    @handle_database_errors
    def execute(self):
        """
        Executes the MySQL command.

        This method utilizes the base class's execute method to
        perform the command execution. It is decorated with
        `handle_database_errors` to handle any MySQL database
        errors that may occur during execution, such as deadlocks
        or conflicts, and raise appropriate exceptions.

        :return: The result of the command execution.
        """
        return super().execute()


class MySQLBatchDelete(base.BaseBatchDelete):
    @handle_database_errors
    def execute(self):
        """
        Executes the MySQL batch delete command.

        This method utilizes the base class's execute method to perform the
        batch delete operation. It is decorated with `handle_database_errors`
        to handle any MySQL database errors that may occur during execution,
        such as deadlocks or conflicts, and raise appropriate exceptions.

        :return: The result of the command execution.
        """

        return super().execute()


class MySQLSelect(base.BaseSelectCommand):
    @handle_database_errors
    def execute(self):
        """
        Executes the MySQL select command.

        This method utilizes the base class's execute method to perform the
        select operation. It is decorated with `handle_database_errors` to
        handle any MySQL database errors that may occur during execution,
        such as deadlocks or conflicts, and raise appropriate exceptions.

        :return: The result of the command execution.
        """
        return super().execute()


class MySQLCustomSelect(base.BaseCustomSelectCommand):
    @handle_database_errors
    def execute(self):
        """
        Executes the MySQL custom select command.

        This method performs the custom select operation by invoking the base
        class's execute method. It is decorated with `handle_database_errors`
        to manage any MySQL database errors, such as deadlocks or conflicts,
        that may arise during the operation, and raises the appropriate
        exceptions.

        :return: The result of the command execution.
        """

        return super().execute()


class MySQLCount(base.BaseCountCommand):
    @handle_database_errors
    def execute(self):
        """
        Executes the MySQL count command.

        This method performs the count operation by invoking the base
        class's execute method. It is decorated with `handle_database_errors`
        to manage any MySQL database errors, such as deadlocks or conflicts,
        that may arise during the operation, and raises the appropriate
        exceptions.

        :return: The result of the command execution.
        """
        return super().execute()


class MySqlOrmDialectCommand(base.BaseOrmDialectCommand):
    @handle_database_errors
    def execute(self):
        """
        Executes the MySQL ORM command.

        This method compiles the query using the `get_statement` method and
        retrieves the values required for the query execution using the
        `get_values` method. Subsequently, it executes the SQL command using
        the parent class's `execute` method, passing the compiled statement
        and values to it. The result of the execution is then wrapped in a
        `BaseOrmProcessResult` object and returned.

        :return: The result of the query execution, wrapped in a
            `BaseOrmProcessResult` object.
        :rtype: BaseOrmProcessResult
        """
        return super().execute()


class MySQLDialect(base.AbstractDialect):
    DIALECT_NAME = "mysql"

    def orm_command(self, table, query, session):
        """
        Creates a new MySQL ORM command for the given table, query, and
        session.

        This method creates an instance of `MySqlOrmDialectCommand` and
        initializes it with the given table, query, and session.

        :param table: The table to be used in the ORM dialect command.
        :param query: The query object that contains the SQL query to be
            compiled.
        :param session: The session to be used for executing the command.
        :return: The created MySQL ORM command.
        :rtype: MySqlOrmDialectCommand
        """
        return MySqlOrmDialectCommand(
            table,
            query,
            session=session,
        )

    def insert(self, table, data, session):
        """
        Inserts data into the specified table using a MySQL dialect command.

        This method creates an instance of `MySQLInsert` to insert the
        provided data into the given table within the specified session
        context.

        :param table: The table into which data is to be inserted.
        :param data: The data to be inserted into the table.
        :param session: The session to be used for executing the insert
            command.
        :return: An instance of `MySQLInsert` configured with the table, data,
            and session.
        :rtype: MySQLInsert
        """

        return MySQLInsert(
            table,
            data,
            session=session,
        )

    def update(self, table, ids, data, session):
        """
        Updates records in the specified table using a MySQL dialect command.

        This method creates an instance of `MySQLUpdate` to update records
        identified by the provided IDs in the given table with the specified
        data within the session context.

        :param table: The table where records are to be updated.
        :param ids: The IDs of the records to be updated.
        :param data: The data to update the records with.
        :param session: The session to be used for executing the update
            command.
        :return: An instance of `MySQLUpdate` configured with the table,
            IDs, data, and session.
        :rtype: MySQLUpdate
        """

        return MySQLUpdate(
            table,
            ids,
            data,
            session=session,
        )

    def delete(self, table, ids, session):
        """
        Deletes records from the specified table using a MySQL dialect
        command.

        This method creates an instance of `MySQLDelete` to delete
        records identified by the provided IDs from the given table within
        the session context.

        :param table: The table from which records are to be deleted.
        :param ids: The IDs of the records to be deleted.
        :param session: The session to be used for executing the delete
            command.
        :return: An instance of `MySQLDelete` configured with the table, IDs,
            and session.
        :rtype: MySQLDelete
        """

        return MySQLDelete(
            table,
            ids,
            session=session,
        )

    def select(self, table, filters, session, limit=None, order_by=None, locked=False):
        """
        Retrieves records from the specified table using a MySQL dialect
        command.

        This method creates an instance of `MySQLSelect` to select records
        from the given table based on the provided filters, with optional
        limit and order by constraints, within the session context.

        :param table: The table from which records are to be selected.
        :param filters: The filters to be used for selecting the records.
        :param session: The session to be used for executing the select
            command.
        :param limit: The maximum number of records to be retrieved.
        :type limit: int, optional
        :param order_by: The ordering criteria for the selected records.
        :param locked: Whether or not to lock the selected records for update.
        :type locked: bool
        :return: An instance of `MySQLSelect` configured with the table,
            filters, session, limit, order by, and locked parameters.
        :rtype: MySQLSelect
        """
        return MySQLSelect(
            table=table,
            filters=filters,
            session=session,
            limit=limit,
            order_by=order_by,
            locked=locked,
        )

    def custom_select(
        self,
        table,
        where_conditions,
        where_values,
        session,
        limit=None,
        order_by=None,
        locked=False,
    ):
        """
        Retrieves records from the specified table using a custom MySQL
        dialect command.

        This method creates an instance of `MySQLCustomSelect` to select
        records from the given table based on the provided where conditions
        and values, with optional limit and order by constraints, within the
        session context.

        :param table: The table from which records are to be selected.
        :param where_conditions: The conditions for the WHERE clause in the
            SQL statement.
        :param where_values: The values for the WHERE clause in the SQL
            statement.
        :type where_values: tuple
        :param session: The session to be used for executing the select
            command.
        :param limit: The maximum number of records to be retrieved.
        :type limit: int, optional
        :param order_by: The ordering criteria for the selected records.
        :param locked: Whether or not to lock the selected records for update.
        :type locked: bool
        :return: An instance of `MySQLCustomSelect` configured with the table,
            where conditions, where values, session, limit, order by, and
            locked parameters.
        :rtype: MySQLCustomSelect
        """
        return MySQLCustomSelect(
            table=table,
            where_conditions=where_conditions,
            where_values=where_values,
            session=session,
            limit=limit,
            order_by=order_by,
            locked=locked,
        )

    def count(self, table, filters, session):
        """
        Retrieves the count of records from the specified table using a
        custom MySQL dialect command.

        This method creates an instance of `MySQLCount` to count records
        from the given table based on the provided filters, with optional
        limit and order by constraints, within the session context.

        :param table: The table from which records are to be counted.
        :param filters: The filters to be used for counting the records.
        :param session: The session to be used for executing the count
            command.
        :return: An instance of `MySQLCount` configured with the table,
            filters, and session parameters.
        :rtype: MySQLCount
        """
        return MySQLCount(
            table=table,
            filters=filters,
            session=session,
        )



# Copyright 2011 OpenStack Foundation.
# Copyright 2020 Eugene Frolov
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webob import dec

from restalchemy.common import contexts as common_contexts

SUCCESS_HTTP_METRIC_NAME = "http.all.success"
ERROR_HTTP_METRIC_NAME = "http.all.errors"


def attach_middlewares(app, middlewares_list):
    for middleware in middlewares_list:
        app = middleware(application=app)
    return app


def configure_middleware(middleware_class, *args, **kwargs):

    def build_middleware(application):
        return middleware_class(application=application, *args, **kwargs)

    return build_middleware


class Middleware(object):
    """Base WSGI middleware wrapper.

    These classes require an application to be initialized that will be called
    next.  By default the middleware will simply call its wrapped app, or you
    can override __call__ to customize its behavior.
    """

    def __init__(self, application):
        self.application = application

    def process_request(self, req):
        """Called on each request.

        If this returns None, the next application down the stack will be
        executed. If it returns a response then that response will be returned
        and execution will stop here.
        """
        return None

    def process_response(self, response):
        """Do whatever you'd like to the response."""
        return response

    @dec.wsgify
    def __call__(self, req):
        response = self.process_request(req)
        if response:
            return response
        response = req.get_response(self.application)
        return self.process_response(response)


class ContextMiddleware(Middleware):
    def process_request(self, req):
        req.context = common_contexts.Context()



# vim: tabstop=4 shiftwidth=4 softtabstop=4
#
# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from wsgiref.simple_server import make_server

from restalchemy.api import applications
from restalchemy.api import controllers
from restalchemy.api import middlewares
from restalchemy.api import resources
from restalchemy.api import routes
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import relationships
from restalchemy.dm import types

HOST = "0.0.0.0"
PORT = 8000


# Storage for foo objects
foo_storage = {}
# Storage for bar objects
bar_storage = {}


# -----------------------------------------------------------------------------
# Models section
# -----------------------------------------------------------------------------
# BarModel 1--->n FooModel


class FooModel(models.ModelWithUUID):
    foo_field1 = properties.property(types.Integer(), required=True)
    foo_field2 = properties.property(types.String(), default="foo_str")


class BarModel(models.ModelWithUUID):
    bar_field1 = properties.property(types.String(min_length=1, max_length=10))
    foo = relationships.relationship(FooModel)


# -----------------------------------------------------------------------------
# Controllers section
# -----------------------------------------------------------------------------


class V1Controller(controllers.Controller):
    """Handle http://127.0.0.1:8000/."""

    def filter(self, **kwargs):
        return ["foos", "bars"]


class FooController(controllers.Controller):
    """Handle http://127.0.0.1:8000/foos/."""

    __resource__ = resources.ResourceByRAModel(FooModel, process_filters=True)

    def create(self, foo_field1, foo_field2):
        foo = self.model(foo_field1=foo_field1, foo_field2=foo_field2)
        foo_storage[str(foo.get_id())] = foo
        return foo

    def get(self, uuid):
        return foo_storage[uuid]

    def filter(self, filters, order_by=None):
        return foo_storage.values()


bar_resource = resources.ResourceByRAModel(BarModel, process_filters=True)


class BarController1(controllers.Controller):
    """Handle http://127.0.0.1:8000/foo/<uuid>/bars/."""

    __resource__ = bar_resource

    def create(self, bar_field1, parent_resource):
        bar = BarModel(bar_field1=bar_field1, foo=parent_resource)
        bar_storage[str(bar.uuid)] = bar
        return bar


class BarController2(controllers.Controller):
    """Handle http://127.0.0.1:8000/bars/<uuid>."""

    __resource__ = bar_resource

    def get(self, uuid):
        return bar_storage[uuid]

    def delete(self, uuid):
        del bar_storage[uuid]


# -----------------------------------------------------------------------------
# Routes section
# -----------------------------------------------------------------------------


class BarRoute1(routes.Route):
    __controller__ = BarController1
    __allow_methods__ = [routes.CREATE]


class BarRoute2(routes.Route):
    __controller__ = BarController2
    __allow_methods__ = [routes.GET, routes.DELETE]


class FooRoute(routes.Route):
    __controller__ = FooController
    __allow_methods__ = [routes.FILTER, routes.CREATE, routes.GET]

    bars = routes.route(BarRoute1, resource_route=True)


class V1Route(routes.Route):
    """Router for / path."""

    __controller__ = V1Controller
    __allow_methods__ = [routes.FILTER]

    # V1Route include two nested routes
    # The first route for /foo/ path
    foos = routes.route(FooRoute)
    bars = routes.route(BarRoute2)


class UserApiApp(routes.Route):
    __controller__ = controllers.RootController
    __allow_methods__ = [routes.FILTER]


# Route to /v1/ endpoint.
setattr(
    UserApiApp,
    "v1",
    routes.route(V1Route),
)


def get_user_api_application():
    return UserApiApp


def build_wsgi_application():
    return middlewares.attach_middlewares(
        applications.WSGIApp(get_user_api_application()),
        [],
    )


def main():
    # Create python WSGI server on any interface with 8000 port
    server = make_server(HOST, PORT, build_wsgi_application())

    try:
        print("Serve forever on %s:%s" % (HOST, PORT))
        server.serve_forever()
    except KeyboardInterrupt:
        print("Bye")


if __name__ == "__main__":
    main()



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import functools

from restalchemy.common import exceptions as common_exc
from restalchemy.common import utils
from restalchemy.storage import exceptions
from restalchemy.storage.sql.dialect import exceptions as dialect_exc


class AbstractObjectCollection(metaclass=abc.ABCMeta):
    def __init__(self, model_cls):
        super(AbstractObjectCollection, self).__init__()
        self.model_cls = model_cls

    @abc.abstractmethod
    def get_all(self, filters=None, limit=None, order_by=None, locked=False):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_one(self, filters=None, locked=False):
        raise NotImplementedError()

    def get_one_or_none(self, filters=None):
        try:
            return self.get_one(filters=filters)
        except exceptions.RecordNotFound:
            return None


class AbstractObjectCollectionCountMixin(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def count(self, filters=None):
        raise NotImplementedError()


class AbstractStorableMixin(metaclass=abc.ABCMeta):
    _ObjectCollection = AbstractObjectCollection

    def _get_prepared_data(self, properties=None):
        result = {}
        props = properties or self.properties
        for name, prop in props.items():
            result[name] = prop.property_type.to_simple_type(prop.value)
        return result

    @utils.classproperty
    def objects(cls):
        return cls._ObjectCollection(cls)

    @abc.abstractmethod
    def insert(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def update(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def save(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def delete(self):
        raise NotImplementedError()

    def get_storable_snapshot(self, properties=None):
        return self._get_prepared_data(properties=properties)


def error_catcher(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except common_exc.RestAlchemyException:
            raise
        except Exception as e:
            raise exceptions.UnknownStorageException(caused=e)

    return wrapper


def dead_lock_catcher(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except dialect_exc.DeadLock as e:
            raise exceptions.DeadLock(msg=str(e))

    return wrapper


class PrefetchResult(dict):
    pass



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.storage.sql import migrations


class MigrationStep(migrations.AbstractMigrationStep):
    def __init__(self):
        self._depends = [""]

    @property
    def migration_id(self):
        return "743d3857-7780-4cdd-8452-2c0aaf316a8b"

    @property
    def is_manual(self):
        return False

    def upgrade(self, session):
        expressions = [
            """
                CREATE TABLE IF NOT EXISTS binary_data (
                    uuid CHAR(36) NOT NULL,
                    data TEXT NOT NULL,
                PRIMARY KEY (uuid))
            """
        ]
        for expression in expressions:
            session.execute(expression)

    def downgrade(self, session):
        tables_to_delete = ["binary_data"]

        for table in tables_to_delete:
            self._delete_table_if_exists(session, table)


migration_step = MigrationStep()



#    Copyright 2021 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest
import webob

from restalchemy.api import constants
from restalchemy.api import contexts
from restalchemy.api import resources
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types


class FakeModel(models.CustomPropertiesMixin, models.ModelWithUUID):
    _private_field = properties.property(types.Integer())
    standard_field1 = properties.property(types.Integer())
    standard_field2 = properties.property(types.Integer())
    standard_field3 = properties.property(types.Integer())
    standard_field4 = properties.property(types.Integer())
    standard_field5 = properties.property(types.Integer())


class FakeMemberContext(object):
    roles = ["member", "some-role"]


class FakeAdminContext(object):
    roles = ["member", "admin"]


class FakeEmptyContext(object):
    roles = []


class FakeIncorrectFieldsRoleContext(object):
    roles = ["incorrect_fields"]


class FakeEmptyFieldsRoleContext(object):
    roles = ["empty_fields"]


# NOTE(efrolov): Interface tests
class ResourceByRAModelHiddenFieldsInterfacesTestCase(unittest.TestCase):
    def tearDown(self):
        super(ResourceByRAModelHiddenFieldsInterfacesTestCase, self).tearDown()
        resources.ResourceMap.model_type_to_resource = {}

    def test_hide_some_fields(self):
        resource = resources.ResourceByRAModel(
            FakeModel,
            hidden_fields=[
                "standard_field1",
                "standard_field4",
                "standard_field5",
            ],
        )

        result = [name for name, prop in resource.get_fields() if prop.is_public()]

        self.assertEqual(["standard_field2", "standard_field3", "uuid"], sorted(result))

    def test_hide_renamed_fields(self):
        resource = resources.ResourceByRAModel(
            FakeModel,
            hidden_fields=[
                "standard_field1",
                "standard_field4",
                "standard_field5",
            ],
            name_map={"standard_field1": "new_standard_field1"},
        )

        result = [name for name, prop in resource.get_fields() if prop.is_public()]

        self.assertEqual(["standard_field2", "standard_field3", "uuid"], sorted(result))


class ResourceByRAModelHiddenFieldsNewInterfacesTestCase(unittest.TestCase):
    def setUp(self):
        super(ResourceByRAModelHiddenFieldsNewInterfacesTestCase, self).setUp()
        self.target = resources.ResourceByRAModel(
            FakeModel,
            hidden_fields=resources.HiddenFieldMap(
                filter=["standard_field1", "standard_field2", "uuid"],
                get=["standard_field1", "standard_field3", "uuid"],
                create=["standard_field1", "standard_field4", "uuid"],
                update=["standard_field1", "standard_field5", "uuid"],
                delete=["standard_field2", "standard_field3", "uuid"],
                action_get=["standard_field2", "standard_field4", "uuid"],
                action_post=["standard_field2", "standard_field5", "uuid"],
                action_put=["standard_field3", "standard_field4", "uuid"],
            ),
        )
        self._request = webob.Request.blank("/some-uri")
        self._request.api_context = contexts.RequestContext(self._request)

    def tearDown(self):
        super(ResourceByRAModelHiddenFieldsNewInterfacesTestCase, self).tearDown()
        resources.ResourceMap.model_type_to_resource = {}
        del self._request

    def _test_hide_some_fields_for_request(self, req, fields):

        result = [
            name
            for name, prop in self.target.get_fields_by_request(req)
            if prop.is_public()
        ]

        self.assertEqual(sorted(fields), sorted(result))

    def test_hide_some_fields_for_filter_method(self):
        self._request.api_context.set_active_method(constants.FILTER)

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=["standard_field3", "standard_field4", "standard_field5"],
        )

    def test_hide_some_fields_for_get_method(self):
        self._request.api_context.set_active_method(constants.GET)

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=["standard_field2", "standard_field4", "standard_field5"],
        )

    def test_hide_some_fields_for_create_method(self):
        self._request.api_context.set_active_method(constants.CREATE)

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=["standard_field2", "standard_field3", "standard_field5"],
        )

    def test_hide_some_fields_for_update_method(self):
        self._request.api_context.set_active_method(constants.UPDATE)

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=["standard_field2", "standard_field3", "standard_field4"],
        )

    def test_hide_some_fields_for_delete_method(self):
        self._request.api_context.set_active_method(constants.DELETE)

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=["standard_field1", "standard_field4", "standard_field5"],
        )

    def test_hide_some_fields_for_action_get_method(self):
        self._request.api_context.set_active_method(constants.ACTION_GET)

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=["standard_field1", "standard_field3", "standard_field5"],
        )

    def test_hide_some_fields_for_action_post_method(self):
        self._request.api_context.set_active_method(constants.ACTION_POST)

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=["standard_field1", "standard_field3", "standard_field4"],
        )

    def test_hide_some_fields_for_action_put_method(self):
        self._request.api_context.set_active_method(constants.ACTION_PUT)

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=["standard_field1", "standard_field2", "standard_field5"],
        )

    def test_get_fields_with_custom_is_public_field_func(self):
        def always_true(name):
            return True

        result = self.target.get_fields(
            override_is_public_field_func=always_true,
        )

        for name, prop in result:
            self.assertTrue(prop.is_public())


class ResourceByRAModelWithCustomPropsHiddenFieldsNewInterfacesTestCase(
    ResourceByRAModelHiddenFieldsNewInterfacesTestCase
):
    def setUp(self):
        self.target = resources.ResourceByModelWithCustomProps(
            FakeModel,
            hidden_fields=resources.HiddenFieldMap(
                filter=["standard_field1", "standard_field2", "uuid"],
                get=["standard_field1", "standard_field3", "uuid"],
                create=["standard_field1", "standard_field4", "uuid"],
                update=["standard_field1", "standard_field5", "uuid"],
                delete=["standard_field2", "standard_field3", "uuid"],
                action_get=["standard_field2", "standard_field4", "uuid"],
                action_post=["standard_field2", "standard_field5", "uuid"],
                action_put=["standard_field3", "standard_field4", "uuid"],
            ),
        )
        self._request = webob.Request.blank("/some-uri")
        self._request.api_context = contexts.RequestContext(self._request)

    def tearDown(self):
        resources.ResourceMap.model_type_to_resource = {}
        del self._request


class ResourceByRAModelRoleBasedHiddenFieldsTestCase(unittest.TestCase):
    def setUp(self):
        super(ResourceByRAModelRoleBasedHiddenFieldsTestCase, self).setUp()
        self.target = resources.ResourceByRAModel(
            FakeModel,
            hidden_fields=resources.RoleBasedHiddenFieldContainer(
                default=resources.HiddenFieldMap(
                    get=[
                        "standard_field1",
                        "standard_field2",
                        "standard_field3",
                        "standard_field4",
                        "standard_field5",
                        "uuid",
                    ],
                ),
                member=resources.HiddenFieldMap(
                    get=["standard_field1", "standard_field3", "uuid"],
                ),
                admin=resources.HiddenFieldMap(
                    get=["standard_field1", "uuid"],
                ),
                incorrect_fields=resources.HiddenFieldMap(
                    get=["fake1", "fake2"],
                ),
                empty_fields=resources.HiddenFieldMap(get=[]),
            ),
        )
        self._request = webob.Request.blank("/some-uri")
        self._request.api_context = contexts.RequestContext(self._request)

    def tearDown(self):
        super(ResourceByRAModelRoleBasedHiddenFieldsTestCase, self).tearDown()
        resources.ResourceMap.model_type_to_resource = {}
        del self._request

    def _test_hide_some_fields_for_request(self, req, fields):

        result = [
            name
            for name, prop in self.target.get_fields_by_request(req)
            if prop.is_public()
        ]

        self.assertEqual(sorted(fields), sorted(result))

    def test_hide_some_fields_for_admin_context_method(self):
        self._request.api_context.set_active_method(constants.GET)
        admin_context = FakeAdminContext()
        self._request.context = admin_context

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=[
                "standard_field2",
                "standard_field3",
                "standard_field4",
                "standard_field5",
            ],
        )

    def test_hide_some_fields_for_member_context_method(self):
        self._request.api_context.set_active_method(constants.GET)
        member_context = FakeMemberContext()
        self._request.context = member_context

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=["standard_field2", "standard_field4", "standard_field5"],
        )

    def test_hide_some_fields_for_empty_role_context_method(self):
        self._request.api_context.set_active_method(constants.GET)
        member_context = FakeEmptyContext()
        self._request.context = member_context

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=[],
        )

    def test_hide_some_fields_without_context(self):
        self._request.api_context.set_active_method(constants.GET)

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=[],
        )

    def test_no_hide_incorrect_fields(self):
        self._request.api_context.set_active_method(constants.GET)
        context = FakeIncorrectFieldsRoleContext()
        self._request.context = context

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=[
                "standard_field1",
                "standard_field2",
                "standard_field3",
                "standard_field4",
                "standard_field5",
                "uuid",
            ],
        )

    def test_no_hide_all_fields(self):
        self._request.api_context.set_active_method(constants.GET)
        context = FakeEmptyFieldsRoleContext()
        self._request.context = context

        self._test_hide_some_fields_for_request(
            req=self._request,
            fields=[
                "standard_field1",
                "standard_field2",
                "standard_field3",
                "standard_field4",
                "standard_field5",
                "uuid",
            ],
        )



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from restalchemy.api import routes
from restalchemy.openapi import structures
from restalchemy.tests.functional.restapi.ra_based.microservice import (
    controllers,
)


class TagRoute(routes.Route):
    __controller__ = controllers.TagController
    __allow_methods__ = [
        routes.CREATE,
        routes.FILTER,
        routes.GET,
        routes.UPDATE,
        routes.DELETE,
    ]


class IpAddress(routes.Route):
    __controller__ = controllers.IpAddressController
    __allow_methods__ = [
        routes.CREATE,
        routes.FILTER,
        routes.GET,
        routes.DELETE,
    ]
    __tags__ = ["IpAddress_tag"]


class PortRoute(routes.Route):
    __controller__ = controllers.PortController
    __allow_methods__ = [
        routes.CREATE,
        routes.FILTER,
        routes.GET,
        routes.UPDATE,
        routes.DELETE,
    ]
    __tags__ = [structures.OpenApiTag(name="PortTestTag", description="port_descr")]
    ip_addresses = routes.route(IpAddress, resource_route=True)


class PortRouteNone(routes.Route):
    __controller__ = controllers.PortControllerNone
    __allow_methods__ = [
        routes.CREATE,
        routes.FILTER,
        routes.GET,
        routes.UPDATE,
        routes.DELETE,
    ]

    ip_addresses = routes.route(IpAddress, resource_route=True)


class NotImplementedAction(routes.Action):
    __controller__ = controllers.NotImplementedMethodsController


class NotImplementedMethodsRoute(routes.Route):
    __controller__ = controllers.NotImplementedMethodsController
    __allow_methods__ = [
        routes.CREATE,
        routes.FILTER,
        routes.GET,
        # routes.UPDATE,  not allowed
        routes.DELETE,
    ]
    not_implemented_action = routes.action(NotImplementedAction, invoke=True)


class VMPowerStateAction(routes.Action):
    __controller__ = controllers.VMController


class VMPowerAction(routes.Action):
    __controller__ = controllers.VMController


class VMPowerOnAction(routes.Action):
    __controller__ = controllers.VMController


class VMPowerOffAction(routes.Action):
    __controller__ = controllers.VMController


class VMRoute(routes.Route):
    __controller__ = controllers.VMController
    __allow_methods__ = [
        routes.CREATE,
        routes.GET,
        routes.DELETE,
        routes.FILTER,
        routes.UPDATE,
    ]

    power = routes.action(VMPowerAction, invoke=True)
    power_state = routes.action(VMPowerStateAction, invoke=False)
    poweron = routes.action(VMPowerOnAction, invoke=True)
    poweroff = routes.action(VMPowerOffAction, invoke=True)
    none_ports = routes.route(PortRouteNone, resource_route=True)
    ports = routes.route(PortRoute, resource_route=True)
    tags = routes.route(TagRoute, resource_route=True)


class VMNoProcessFiltersRoute(routes.Route):
    __controller__ = controllers.VMNoProcessFiltersController
    __allow_methods__ = [
        routes.GET,
        routes.FILTER,
    ]


class VMNoSortRoute(routes.Route):
    __controller__ = controllers.VMNoSortController
    __allow_methods__ = [
        routes.CREATE,
        routes.GET,
        routes.DELETE,
        routes.FILTER,
        routes.UPDATE,
    ]


class VMDefSortRoute(routes.Route):
    __controller__ = controllers.VMDefSortController
    __allow_methods__ = [
        routes.CREATE,
        routes.GET,
        routes.DELETE,
        routes.FILTER,
        routes.UPDATE,
    ]


class V1Route(routes.Route):
    __controller__ = controllers.V1Controller
    __allow_methods__ = [routes.FILTER]

    vms = routes.route(VMRoute)
    vmsnoproccessfilters = routes.route(VMNoProcessFiltersRoute)
    vmsnosort = routes.route(VMNoSortRoute)
    vmsdefsort = routes.route(VMDefSortRoute)
    notimplementedmethods = routes.route(NotImplementedMethodsRoute)


class Root(routes.RootRoute):
    v1 = routes.route(V1Route)



# Copyright 2022 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.openapi import constants
from restalchemy.openapi import structures


class OpenApi303(object):
    """OpenAPI Specification

    The OpenAPI Specification (OAS) defines a standard, language-agnostic
    interface to RESTful APIs which allows both humans and computers to
    discover and understand the capabilities of the service without access to
    source code, documentation, or through network traffic inspection. When
    properly defined, a consumer can understand and interact with the remote
    service with a minimal amount of implementation logic.

    An OpenAPI definition can then be used by documentation generation tools
    to display the API, code generation tools to generate servers and clients
    in various programming languages, testing tools, and many other use cases.

    :param info: REQUIRED. Provides metadata about the API. The metadata MAY
        be used by tooling as required.
    :type info: :class:`OpenApiInfo`
    :param paths: The available paths and operations for the API.
    :type paths: :class:`OpenApiPaths`
    :param servers: An array of Server Objects, which provide connectivity
        information to a target server. If the servers property is not
        provided, or is an empty array, the default value would be a Server
        Object with a url value of /.
    :param components: An element to hold various schemas for the
        specification.
    :type components: :class:`OpenApiComponents`
    :param security: A declaration of which security mechanisms can be used
        across the API. The list of values includes alternative security
        requirement objects that can be used. Only one of the security
        requirement objects need to be satisfied to authorize a request.
        Individual operations can override this definition. To make security
        optional, an empty security requirement ({}) can be included in the
        array.
    :param tags: A list of tags used by the specification with additional
        metadata. The order of the tags can be used to reflect on their order
        by the parsing tools. Not all tags that are used by the Operation
        Object must be declared. The tags that are not declared MAY be
        organized randomly or based on the tools' logic. Each tag name in the
        list MUST be unique.
    :param external_docs: Additional external documentation.
    """

    def __init__(
        self,
        info,
        paths=None,
        servers=None,
        components=None,
        security=None,
        tags=None,
        external_docs=None,
    ):
        super(OpenApi303, self).__init__()
        self._info = info or structures.OpenApiInfo()
        self._paths = paths or structures.OpenApiPaths()
        self._servers = servers or structures.OpenApiServers()
        self._components = components or structures.OpenApiComponents()
        self._security = security
        self._tags = tags or structures.OpenApiTags([])
        self._external_docs = external_docs

    def build_openapi_specification(self, request):
        specification = {
            "openapi": constants.OPENAPI_SPECIFICATION_3_0_3,
        }

        specification.update(self._info.build(request))
        components = self._components.build(request)
        specification.update(components)
        specification.update(self._paths.build(request, components))
        specification.update(self._servers.build(request))
        specification.update(self._tags.build(request))
        return specification



# Copyright 2022 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

LOG = logging.getLogger(__name__)


class ResourceSchemaGenerator(object):
    def __init__(self, resource, route):
        super(ResourceSchemaGenerator, self).__init__()
        self._resource = resource
        self._route = route

    @property
    def resource_name(self):
        return self._resource.get_model().__name__

    def resource_method_name(self, method):
        return "{}_{}".format(self.resource_name, method.capitalize())

    def resource_prop_name(self, prop_name):
        return self.resource_name + prop_name.capitalize()

    def get_prop_kwargs(self, name):
        return self._resource.get_model().properties.properties[name].get_kwargs()

    def generate_parameter_object(self, request):
        parameters = {}
        has_id_property = False
        for name, prop in self._resource.get_fields_by_request(request):
            try:
                prop_kwargs = self.get_prop_kwargs(name)
            except KeyError:
                prop_kwargs = {}
            schema = prop.get_type().to_openapi_spec(prop_kwargs)
            try:
                is_id = prop.is_id_property()
            except KeyError:
                is_id = False
            if is_id:
                has_id_property = True
                prop_name = self.resource_prop_name(name)
            else:
                prop_name = prop.api_name
            parameters[prop_name] = {
                "name": prop_name,
                "in": "path" if is_id else "query",
                "schema": schema,
            }
            if is_id:
                parameters[prop_name]["required"] = True
        if not has_id_property:
            try:
                model = self._resource.get_model()
                id_prop_struct = model.get_id_property()
                id_prop = list(id_prop_struct.items())[0]
                name, prop = id_prop
                prop = prop(value=prop._kwargs.get("default", 0))
                try:
                    prop_kwargs = self.get_prop_kwargs(name)
                except KeyError:
                    prop_kwargs = {}
                schema = prop.get_property_type().to_openapi_spec(prop_kwargs)
                prop_name = self.resource_prop_name(name)
                parameters[prop_name] = {
                    "name": prop_name,
                    "in": "path",
                    "schema": schema,
                }
                parameters[prop_name]["required"] = True
            except Exception:
                LOG.exception("Error on generate_parameter_object:")
        return parameters

    def generate_schema_object(self, method):
        return self._resource.generate_schema_object(method)


class Schema(object):
    def __init__(
        self,
        summary=None,
        parameters=None,
        responses=None,
        tags=None,
        request_body=None,
        operation_id=None,
    ):
        self.summary = summary or ""
        self.parameters = parameters or []
        self.responses = responses or {}
        self.tags = tags or []
        self.request_body = request_body
        self.operation_id = operation_id

    @property
    def result(self):
        res = {
            "summary": self.summary,
            "tags": self.tags,
            "parameters": self.parameters,
            "responses": self.responses,
        }
        if self.request_body is not None:
            res["requestBody"] = self.request_body
        if self.operation_id is not None:
            res["operationId"] = self.operation_id
        return res


def extend_schema(
    summary=None,
    parameters=None,
    responses=None,
    tags=None,
    request_body=None,
    operation_id=None,
):
    if parameters and not isinstance(parameters, list):
        raise ValueError("parameters type is not list")
    if responses and not isinstance(responses, dict):
        raise ValueError("responses type is not dict")
    if tags and not isinstance(tags, list):
        raise ValueError("tags type is not list")

    def decorator(f):
        schema = Schema(
            summary=summary,
            parameters=parameters,
            responses=responses,
            tags=tags,
            request_body=request_body,
            operation_id=operation_id,
        )
        f.openapi_schema = schema
        return f

    return decorator



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.common import contexts
from restalchemy.storage.sql import engines


class Context(contexts.Context):
    _engine = None
    _session = None

    def get_session(self):
        if self._engine is None:
            self._engine = engines.engine_factory.get_engine()
        return self._engine.get_session()

    @property
    def session(self):
        if self._session is None:
            self._session = self.get_session()
        return self._session

    def release(self):
        if self._session:
            self._session.close()
            self._session = None



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
# Copyright 2019 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

import mock

from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import relationships
from restalchemy.dm import types
from restalchemy.common import contexts
from restalchemy.storage import exceptions
from restalchemy.storage.sql import engines
from restalchemy.storage.sql import orm
from restalchemy.storage.sql import utils
from restalchemy.tests.functional import base
from restalchemy.tests.functional import consts
from restalchemy.tests import fixtures

FAKE_STR1 = "Fake1"
FAKE_STR2 = "Fake2"
FAKE_INT1 = 1
FAKE_INT2 = 2
FAKE_URI1 = "/fake/00000000-0000-0000-0000-700000000001"
FAKE_URI2 = "/fake/00000000-0000-0000-0000-700000000002"
FAKE_MAC1 = "00:01:02:03:04:05"
FAKE_MAC2 = "10:11:12:13:14:15"
FAKE_TABLE_NAME1 = "Fake-table-name1"
FAKE_TABLE_NAME2 = "Fake-table-name2"
FAKE_UUID1 = uuid.UUID("00000000-0000-0000-0000-700000000003")
FAKE_UUID2 = uuid.UUID("00000000-0000-0000-0000-700000000004")
FAKE_UUID1_STR = str(FAKE_UUID1)
FAKE_UUID2_STR = str(FAKE_UUID2)

URL_TO_DB = "mysql://fake:fake@127.0.0.1/test"


class FakeParentModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = FAKE_TABLE_NAME2


class FakeModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = FAKE_TABLE_NAME1

    test_str_field1 = properties.property(types.String(), default=FAKE_STR1)
    test_str_field2 = properties.property(types.String(), default=FAKE_STR2)

    test_int_field1 = properties.property(types.Integer(), default=FAKE_INT1)
    test_int_field2 = properties.property(types.Integer(), default=FAKE_INT2)

    test_uri_field1 = properties.property(types.Uri(), default=FAKE_URI1)
    test_uri_field2 = properties.property(types.Uri(), default=FAKE_URI2)

    test_mac_field1 = properties.property(types.Mac(), default=FAKE_MAC1)
    test_mac_field2 = properties.property(types.Mac(), default=FAKE_MAC2)

    test_parent_relationship = relationships.relationship(FakeParentModel)


class FakeModelWithValidate(FakeModel):
    def validate(self):
        if self.test_str_field1 == self.test_str_field2:
            raise ValueError


ROW = {
    "test_str_field1": FAKE_STR1,
    "test_str_field2": FAKE_STR2,
    "test_int_field1": FAKE_INT1,
    "test_int_field2": FAKE_INT2,
    "test_uri_field1": FAKE_URI1,
    "test_uri_field2": FAKE_URI2,
    "test_mac_field1": FAKE_MAC1,
    "test_mac_field2": FAKE_MAC2,
    "uuid": FAKE_UUID1_STR,
    "test_parent_relationship": FAKE_UUID2_STR,
}


COLUMNS_NAME = sorted(ROW.keys())
VALUES = tuple()
for key in sorted(ROW.keys()):
    VALUES += (ROW[key],)


def escape(list_of_fields):
    return ["`%s`" % field for field in list_of_fields]


class InsertCaseTestCase(base.BaseFunctionalTestCase):
    @mock.patch("mysql.connector.pooling.MySQLConnectionPool")
    def setUp(self, mysql_pool_mock):
        super(InsertCaseTestCase, self).setUp()

        # configure engine factory
        engines.engine_factory.configure_factory(db_url=URL_TO_DB)
        self.parent_model = FakeParentModel(uuid=FAKE_UUID2)
        self.target_model = FakeModel(
            uuid=FAKE_UUID1, test_parent_relationship=self.parent_model
        )

    def tearDown(self):
        super(InsertCaseTestCase, self).tearDown()

        del self.target_model
        # Note(efrolov): Must be deleted otherwise we will start collect
        #                connections and get an error "too many connections"
        #                from MySQL
        engines.engine_factory.destroy_engine()

    @mock.patch(
        "restalchemy.storage.sql.sessions.MySQLSession",
        return_value=fixtures.SessionFixture(),
    )
    def test_insert_new_model_session_is_none(self, session_mock):

        self.target_model.insert()

        session_mock().execute.assert_called_once_with(
            "INSERT INTO `%s` (%s) VALUES (%s)"
            % (
                FAKE_TABLE_NAME1,
                ", ".join(escape(COLUMNS_NAME)),
                ", ".join(["%s"] * len(VALUES)),
            ),
            VALUES,
        )
        self.assertTrue(session_mock().commit.called)
        self.assertFalse(session_mock().rollback.called)
        self.assertTrue(session_mock().close.called)

    @mock.patch(
        "restalchemy.storage.sql.sessions.MySQLSession",
        return_value=fixtures.SessionFixture(),
    )
    def test_insert_new_model_session_is_none_and_db_error(self, session_mock):

        class CustomException(Exception):
            pass

        session_mock().execute.side_effect = CustomException

        self.assertRaises(exceptions.UnknownStorageException, self.target_model.insert)

        self.assertFalse(session_mock().commit.called)
        self.assertTrue(session_mock().rollback.called)
        self.assertTrue(session_mock().close.called)

    def test_insert_new_model_with_session(self):
        session_mock = fixtures.SessionFixture()

        self.target_model.insert(session=session_mock)

        session_mock.execute.assert_called_once_with(
            "INSERT INTO `%s` (%s) VALUES (%s)"
            % (
                FAKE_TABLE_NAME1,
                ", ".join(escape(COLUMNS_NAME)),
                ", ".join(["%s"] * len(VALUES)),
            ),
            VALUES,
        )
        self.assertFalse(session_mock.commit.called)
        self.assertFalse(session_mock.rollback.called)
        self.assertFalse(session_mock.close.called)

    def test_insert_new_model_with_session_and_db_error(self):

        session_mock = fixtures.SessionFixture()

        class CustomException(Exception):
            pass

        session_mock.execute.side_effect = CustomException

        self.assertRaises(
            exceptions.UnknownStorageException,
            self.target_model.insert,
            session=session_mock,
        )

        self.assertFalse(session_mock.commit.called)
        self.assertFalse(session_mock.rollback.called)
        self.assertFalse(session_mock.close.called)

    def test_validate(self):
        with self.assertRaises(ValueError):
            FakeModelWithValidate(
                uuid=FAKE_UUID1,
                test_parent_relationship=self.parent_model,
                test_str_field1=FAKE_STR1,
                test_str_field2=FAKE_STR1,
            )


class FakeUpdateModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "test_update"

    field1 = properties.property(types.String(), required=True)
    field2 = properties.property(types.String(), required=True)


class UpdateTestCase(base.BaseFunctionalTestCase):
    def setUp(self):
        super(UpdateTestCase, self).setUp()

        engines.engine_factory.configure_factory(consts.get_database_uri())
        engine = engines.engine_factory.get_engine()
        self.engine = engine
        with engine.session_manager() as session:
            session.execute("""
                CREATE TABLE IF NOT EXISTS test_update (
                    uuid CHAR(36) PRIMARY KEY,
                    field1 VARCHAR(255) NOT NULL,
                    field2 VARCHAR(255) NOT NULL
                )
            """)

    def tearDown(self):
        super(UpdateTestCase, self).tearDown()

        with self.engine.session_manager() as session:
            session.execute("DROP TABLE IF EXISTS test_update;", None)
            # Note(efrolov): Must be deleted otherwise we will start collect
            #                connections and get an error "too many connections"
            #                from MySQL
        engines.engine_factory.destroy_engine()

    def test_update_not_changed_model(self):
        test_model = FakeUpdateModel(field1=FAKE_STR1, field2=FAKE_STR2)
        test_model.save()

        self.assertIsNone(test_model.update())

    def test_force_update_not_changed_model(self):
        test_model = FakeUpdateModel(field1=FAKE_STR1, field2=FAKE_STR2)
        test_model.save()

        self.assertIsNone(test_model.update(force=True))

    def test_validate(self):
        model = FakeModelWithValidate(
            uuid=FAKE_UUID1,
            test_parent_relationship=FakeParentModel(uuid=FAKE_UUID2),
        )
        model.test_str_field2 = FAKE_STR1
        with self.assertRaises(exceptions.UnknownStorageException):
            model.update()


class SavepointModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "savepoint_table"

    field1 = properties.property(types.String(), required=True)
    field2 = properties.property(types.String(), required=True)


class SavepointTestCase(base.BaseFunctionalTestCase):
    def setUp(self):
        super().setUp()

        engines.engine_factory.configure_factory(consts.get_database_uri())
        engine = engines.engine_factory.get_engine()
        self.engine = engine
        with engine.session_manager() as session:
            session.execute("""
                CREATE TABLE IF NOT EXISTS savepoint_table (
                    uuid CHAR(36) PRIMARY KEY,
                    field1 VARCHAR(255) NOT NULL,
                    field2 VARCHAR(255) NOT NULL
                )
            """)

    def tearDown(self):
        super().tearDown()

        with self.engine.session_manager() as session:
            session.execute("DROP TABLE IF EXISTS savepoint_table;", None)
        engines.engine_factory.destroy_engine()

    def test_savepoint_success_result(self):
        with contexts.Context().session_manager():
            test_model = SavepointModel(field1=FAKE_STR1, field2=FAKE_STR2)

            with utils.savepoint():
                test_model.save()

            self.assertEqual(test_model.field1, FAKE_STR1)
            self.assertEqual(test_model.field2, FAKE_STR2)

    def test_savepoint_rollback_result(self):
        with contexts.Context().session_manager():
            test_model = SavepointModel(field1=FAKE_STR1, field2="")
            test_model.save()

            self.assertEqual(test_model.field1, FAKE_STR1)
            self.assertEqual(test_model.field2, "")

            def save_and_raise():
                test_model.field2 = FAKE_STR2
                test_model.save()
                raise ValueError("Error")

            with self.assertRaises(ValueError):
                with utils.savepoint():
                    save_and_raise()

        with contexts.Context().session_manager():
            objects = SavepointModel.objects.get_all()

            self.assertEqual(len(objects), 1)
            self.assertEqual(objects[0].field1, FAKE_STR1)
            self.assertEqual(objects[0].field2, "")

    def test_savepoint_can_continue(self):
        with contexts.Context().session_manager():
            test_model = SavepointModel(field1=FAKE_STR1, field2="")
            test_model.save()

            self.assertEqual(test_model.field1, FAKE_STR1)
            self.assertEqual(test_model.field2, "")

            def save_and_raise():
                test_model.field2 = FAKE_STR2
                test_model.save()
                raise ValueError("Error")

            with self.assertRaises(ValueError):
                with utils.savepoint():
                    save_and_raise()

            test_model.field2 = "foo-value"
            test_model.save()

        with contexts.Context().session_manager():
            objects = SavepointModel.objects.get_all()

            self.assertEqual(len(objects), 1)
            self.assertEqual(objects[0].field1, FAKE_STR1)
            self.assertEqual(objects[0].field2, "foo-value")



# Copyright 2019 Eugene Frolov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import absolute_import

import logging
import re
import time

from restalchemy.api import middlewares

LOG = logging.getLogger(__name__)


class HttpMetricsMiddleware(middlewares.Middleware):
    def __init__(
        self,
        application,
        path_pattern,
        success_metric_name,
        error_metric_name,
        metric_sender,
    ):
        super(HttpMetricsMiddleware, self).__init__(application)
        self._re = re.compile(path_pattern)
        self._success_metric_name = success_metric_name
        self._error_metric_name = error_metric_name
        self._sender = metric_sender

    def process_request(self, req):
        if self._re.match(req.path) is None:
            return req.get_response(self.application)

        current_time = time.time()
        try:
            res = req.get_response(self.application)
            elapsed = time.time() - current_time
            if res.status_code >= 400:
                self._sender.send_metric(self._error_metric_name, elapsed)
                self._sender.send_metric(
                    "%s.%d" % (self._error_metric_name, res.status_code),
                    elapsed,
                )
            else:
                self._sender.send_metric(self._success_metric_name, elapsed)
            return res
        except Exception:
            self._sender.send_metric(
                "%s.unexpected-error" % self._error_metric_name,
                time.time() - current_time,
            )
            raise



#    Copyright 2021 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from parameterized import parameterized

from restalchemy.dm import filters as dm_filters
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.storage.sql import engines
from restalchemy.storage.sql import orm
from restalchemy.tests.functional import base
from restalchemy.tests.functional import consts
from restalchemy.tests.utils import make_test_name


class FakeModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "batch_insert"
    foo_field1 = properties.property(types.Integer(), required=True)
    foo_field2 = properties.property(types.String())


class TestOrderByTestCase(base.BaseWithDbMigrationsTestCase):
    __LAST_MIGRATION__ = "test-batch-migration-9e335f"
    __FIRST_MIGRATION__ = "test-batch-migration-9e335f"

    def test_without_order_by(self):
        model1 = FakeModel(foo_field1=1, foo_field2="Model1")
        model2 = FakeModel(foo_field1=2, foo_field2="Model2")

        with self.engine.session_manager() as session:
            session.batch_insert([model1, model2])

        all_models = set(FakeModel.objects.get_all())

        self.assertEqual({model1, model2}, all_models)

    @parameterized.expand(
        [
            (None, ("1", "2")),
            ("ASC", ("1", "2")),
            ("DESC", ("2", "1")),
            ("ASC NULLS FIRST", (None, "1", "2")),
            ("ASC NULLS LAST", ("1", "2", None)),
            ("DESC NULLS FIRST", (None, "2", "1")),
            ("DESC NULLS LAST", ("2", "1", None)),
        ],
        name_func=make_test_name,
    )
    def test_with_various_order_bys(self, sort_dir, correct_order):
        items = [
            FakeModel(foo_field1=idx, foo_field2=val)
            for idx, val in enumerate(correct_order)
        ]
        with self.engine.session_manager() as session:
            session.batch_insert(items)

        results_from_db = FakeModel.objects.get_all(order_by={"foo_field2": sort_dir})
        values_from_db = tuple(i.foo_field2 for i in results_from_db)
        self.assertEqual(correct_order, values_from_db)


class TestLikeTestCase(base.BaseWithDbMigrationsTestCase):
    __LAST_MIGRATION__ = "test-batch-migration-9e335f"
    __FIRST_MIGRATION__ = "test-batch-migration-9e335f"

    def test_filter_like_not_like(self):
        model1 = FakeModel(foo_field1=1, foo_field2="Model1")
        model2 = FakeModel(foo_field1=2, foo_field2="Model2")
        model3 = FakeModel(foo_field1=3, foo_field2="FooModel3")
        model4 = FakeModel(foo_field1=4, foo_field2="TestMod-4")
        model5 = FakeModel(foo_field1=5, foo_field2="Mod'el")

        with self.engine.session_manager() as session:
            session.batch_insert([model1, model2, model3, model4, model5])

        # Like: Startswith Model
        all_models = set(
            FakeModel.objects.get_all(filters={"foo_field2": dm_filters.Like("Model%")})
        )

        self.assertEqual({model1, model2}, all_models)

        # Like: del in center
        all_models = set(
            FakeModel.objects.get_all(filters={"foo_field2": dm_filters.Like("%del%")})
        )

        self.assertEqual({model1, model2, model3}, all_models)

        # Like: Endswith 3
        all_models = set(
            FakeModel.objects.get_all(filters={"foo_field2": dm_filters.Like("%3")})
        )

        self.assertEqual(
            {
                model3,
            },
            all_models,
        )

        # Like: TestMod + random symbol + 4
        all_models = set(
            FakeModel.objects.get_all(
                filters={"foo_field2": dm_filters.Like("TestMod_4")}
            )
        )

        self.assertEqual(
            {
                model4,
            },
            all_models,
        )

        # Like: Endswith 5
        all_models = set(
            FakeModel.objects.get_all(filters={"foo_field2": dm_filters.Like("%5")})
        )

        self.assertEqual(set(), all_models)

        # Like: Startswith Foo or Test
        # TODO(v.burygin) Зачем тут AND, - неизвестно, надо разбираться
        filter_list = dm_filters.AND(
            dm_filters.OR(
                {"foo_field2": dm_filters.Like("Foo%")},
                {"foo_field2": dm_filters.Like("Test%")},
            )
        )
        all_models = set(FakeModel.objects.get_all(filters=filter_list))

        self.assertEqual({model3, model4}, all_models)

        # Like: ' in center with escape symbol
        all_models = set(
            FakeModel.objects.get_all(
                filters={"foo_field2": dm_filters.Like(r"Mod\'el")}
            )
        )

        self.assertEqual(
            {
                model5,
            },
            all_models,
        )

        # NotLike: Not startswith Foo
        all_models = set(
            FakeModel.objects.get_all(
                filters={"foo_field2": dm_filters.NotLike("Foo%")}
            )
        )

        self.assertEqual({model1, model2, model4, model5}, all_models)

        # NotLike: Mod in center
        all_models = set(
            FakeModel.objects.get_all(
                filters={"foo_field2": dm_filters.NotLike("%Mod%")}
            )
        )

        self.assertEqual(set(), all_models)

        # NotLike: ' in center with escape symbol
        all_models = set(
            FakeModel.objects.get_all(
                filters={"foo_field2": dm_filters.NotLike(r"Mod\'el")}
            )
        )

        self.assertEqual({model1, model2, model3, model4}, all_models)


class TestCacheTestCase(base.BaseWithDbMigrationsTestCase):
    __LAST_MIGRATION__ = "test-batch-migration-9e335f"
    __FIRST_MIGRATION__ = "test-batch-migration-9e335f"

    @classmethod
    def init_engine(cls):
        engines.engine_factory.configure_factory(
            db_url=consts.get_database_uri(), query_cache=True
        )
        cls.__ENGINE__ = engines.engine_factory.get_engine()

    def test_get_one_all_with_cache(self):
        model1 = FakeModel(foo_field1=1, foo_field2="Model1")

        with self.engine.session_manager() as session:
            session.batch_insert([model1])

            tgt_model = FakeModel.objects.get_one(session=session, cache=True)
            tgt_model_cached = FakeModel.objects.get_one(session=session, cache=True)

            assert tgt_model is tgt_model_cached

    def test_get_one_all_without_cache(self):
        model1 = FakeModel(foo_field1=1, foo_field2="Model1")

        with self.engine.session_manager() as session:
            session.batch_insert([model1])

            tgt_model = FakeModel.objects.get_one(session=session, cache=True)
            tgt_model_cached = FakeModel.objects.get_one(session=session, cache=False)

            assert tgt_model is not tgt_model_cached



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.storage.sql import migrations


class MigrationStep(migrations.AbstractMigrationStep):
    def __init__(self):
        self._depends = [""]

    @property
    def migration_id(self):
        return "e31a12bb-3c3a-4f86-8bdd-ca9f7b613b6c"

    def upgrade(self, session):
        expressions = [
            """
                CREATE TABLE IF NOT EXISTS vms (
                    uuid CHAR(36) NOT NULL,
                    state VARCHAR(10) NOT NULL,
                    name VARCHAR(255) NOT NULL,
                    just_none VARCHAR(255) NULL,
                    status VARCHAR(255) NULL,
                    created TIMESTAMP(6) DEFAULT CURRENT_TIMESTAMP,
                    updated TIMESTAMP(6) DEFAULT CURRENT_TIMESTAMP,
                    PRIMARY KEY (uuid)
                );
            """,
            """
                CREATE TABLE IF NOT EXISTS ports (
                    uuid CHAR(36) NOT NULL,
                    mac CHAR(17) NOT NULL,
                    vm CHAR(36) NOT NULL,
                    PRIMARY KEY (uuid),
                    FOREIGN KEY (vm) REFERENCES vms (uuid)
                );
            """,
            """
                CREATE TABLE IF NOT EXISTS ip_addresses (
                    uuid CHAR(36) NOT NULL,
                    ip VARCHAR(17) NOT NULL,
                    port CHAR(36) NOT NULL,
                    PRIMARY KEY (uuid),
                    FOREIGN KEY (port) REFERENCES ports (uuid)
                );
            """,
            """
                CREATE TABLE IF NOT EXISTS tags (
                    uuid CHAR(36) NOT NULL,
                    vm CHAR(36) NOT NULL,
                    name VARCHAR(40) NOT NULL,
                    visible BOOLEAN NOT NULL,
                    PRIMARY KEY (uuid),
                    FOREIGN KEY (vm) REFERENCES vms (uuid)
                );
            """,
        ]

        for expression in expressions:
            session.execute(expression)

    def downgrade(self, session):
        tables = ["tags", "ip_addresses", "ports", "vms"]

        for table in tables:
            self._delete_table_if_exists(session, table)


migration_step = MigrationStep()



#    Copyright 2020 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


OPERATIONAL_STORAGE_SIMPLE_TABLE_KEY = "table"


class SQLTable(object):
    def __init__(self, engine, table_name, model):
        super(SQLTable, self).__init__()
        self._table_name = table_name
        self._model = model

    @property
    def model(self):
        return self._model

    def get_column_names(self, session, with_pk=True, do_sort=True):
        result = []
        for name, prop in self._model.properties.items():
            if not with_pk and prop.is_id_property():
                continue
            result.append(name)
        if do_sort:
            result.sort()
        return result

    def get_escaped_column_names(self, session, with_pk=True, do_sort=True):
        return [
            session.engine.escape(column_name)
            for column_name in self.get_column_names(
                session=session,
                with_pk=with_pk,
                do_sort=do_sort,
            )
        ]

    def get_pk_names(self, session, do_sort=True):
        result = []
        for name, prop in self._model.properties.items():
            if prop.is_id_property():
                result.append(name)
        if do_sort:
            result.sort()
        return result

    def get_escaped_pk_names(self, session, do_sort=True):
        return [
            session.engine.escape(column_name)
            for column_name in self.get_pk_names(
                session=session,
                do_sort=do_sort,
            )
        ]

    @property
    def name(self):
        return self._table_name

    def insert(self, engine, data, session):
        cmd = engine.dialect.insert(
            table=self,
            data=data,
            session=session,
        )
        return cmd.execute()

    def update(self, engine, ids, data, session):
        cmd = engine.dialect.update(
            table=self,
            ids=ids,
            data=data,
            session=session,
        )
        return cmd.execute()

    def delete(self, engine, ids, session):
        cmd = engine.dialect.delete(
            table=self,
            ids=ids,
            session=session,
        )
        return cmd.execute()

    def select(self, engine, filters, session, limit=None, order_by=None, locked=False):
        """

        Warning: query with and w/o (limit or group_by) won't flush each other
        if cached!
        """
        q = engine.dialect.orm.select(self._model, session).where(
            filters=filters,
        )

        for name, sort_type in (order_by or {}).items():
            q.order_by(property_name=name, sort_type=sort_type)

        if limit:
            q.limit(limit)

        if locked:
            q.for_(share=not locked)

        cmd = engine.dialect.orm_command(
            table=self,
            query=q,
            session=session,
        )
        return cmd.execute()

    def custom_select(
        self,
        engine,
        where_conditions,
        where_values,
        session,
        limit=None,
        order_by=None,
        locked=False,
    ):
        cmd = engine.dialect.custom_select(
            table=self,
            where_conditions=where_conditions,
            where_values=where_values,
            limit=limit,
            order_by=order_by,
            locked=locked,
            session=session,
        )
        return cmd.execute()

    def count(self, engine, session, filters):
        cmd = engine.dialect.count(
            table=self,
            filters=filters,
            session=session,
        )
        return cmd.execute()



# Copyright 2023 George Melikov
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os
import sys

from oslo_config import cfg

from restalchemy.common import config
from restalchemy.common import log as ra_log
from restalchemy.storage.sql import migrations


def suggest_filename(file, migration_files):
    index = "{:04d}".format(migration_files["index"])
    filename = file.rstrip(".py")
    uuid = migration_files["uuid"].split("-")[0][:6]

    if not migration_files["is_manual"]:
        result = "{}-{}-{}.py".format(index, filename, uuid)
    else:
        result = "MANUAL-{}-{}.py".format(filename, uuid)

    return result


cmd_opts = [
    cfg.StrOpt(
        "path",
        required=True,
        short="p",
        help="Path to migrations folder",
    ),
]

CONF = cfg.CONF
CONF.register_cli_opts(cmd_opts)


def main():
    config.parse(sys.argv[1:])
    ra_log.configure()

    engine = migrations.MigrationEngine(migrations_path=CONF.path)

    migration_files = engine.get_all_migrations()

    logging.info("Parsing files in folder - %s", CONF.path)

    for file in migration_files.keys():
        suggested_name = suggest_filename(file, migration_files[file])

        logging.info("Renaming %s to %s", file, suggested_name)

        os.rename(
            os.path.join(CONF.path, file),
            os.path.join(CONF.path, suggested_name),
        )

        if not migration_files[file]["depends"]:
            continue

        with open(os.path.join(CONF.path, suggested_name), "r+") as f:
            content = f.read()

            for d in migration_files[file]["depends"]:
                suggested_depends_name = suggest_filename(
                    d,
                    migration_files[d],
                )
                logging.info(
                    "Renaming depends %s to %s",
                    d,
                    suggested_depends_name,
                )
                content = content.replace(d, suggested_depends_name)

            f.seek(0)

            f.write(content)

            f.truncate()


if __name__ == "__main__":
    main()



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.common import utils
from restalchemy.tests.unit import base


class ReadOnlyDictProxyTestCase(base.BaseTestCase):
    def test_is(self):
        self.assertFalse(
            utils.ReadOnlyDictProxy({1: 2}) is utils.ReadOnlyDictProxy({1: 2})
        )

    def test_equal(self):
        self.assertEqual(
            utils.ReadOnlyDictProxy({1: 2}), utils.ReadOnlyDictProxy({1: 2})
        )

    def test_len(self):
        self.assertEqual(len(utils.ReadOnlyDictProxy({1: 2, 2: 4})), 2)

    def test_getitem(self):
        self.assertEqual(utils.ReadOnlyDictProxy({1: 2, 2: 4})[1], 2)

    def test_hash(self):
        hash1 = hash(utils.ReadOnlyDictProxy({1: 2, 2: 4}))
        hash2 = hash(utils.ReadOnlyDictProxy({1: 2, 2: 4}))
        hash3 = hash(utils.ReadOnlyDictProxy({1: 2, 2: 3}))

        self.assertEqual(hash1, hash2)
        self.assertNotEqual(hash2, hash3)

    def test_readonly(self):
        dict1 = utils.ReadOnlyDictProxy({1: 2, 2: 4})

        def rotest(d):
            d[1] = 5

        self.assertRaises(TypeError, rotest, dict1)



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

import mock

from restalchemy.common import exceptions
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import relationships
from restalchemy.dm import types
from restalchemy.tests.unit import base


class FakeProperty(mock.Mock):
    pass


FAKE_PROPERTY1 = FakeProperty()
FAKE_PROPERTY2 = FakeProperty()
FAKE_VALUE1 = "Fake value 1"


class MetaModelTestCase(base.BaseTestCase):
    with mock.patch("restalchemy.dm.properties.PropertyCreator", FakeProperty):

        class Model(object, metaclass=models.MetaModel):
            fake_prop1 = FAKE_PROPERTY1
            fake_prop2 = FAKE_PROPERTY2

    def test_model_class(self):
        self.assertIsInstance(self.Model.properties, properties.PropertyCollection)

    def test_metamodel_getattr(self):
        self.assertNotEqual(self.Model.fake_prop1, FAKE_PROPERTY1)
        self.assertNotEqual(self.Model.fake_prop2, FAKE_PROPERTY2)

    def test_metamodel_getattr_raises_attribute_error(self):
        self.assertRaises(AttributeError, lambda: self.Model.fake_prop3)


class ModelTestCase(base.BaseTestCase):
    PM_MOCK = mock.MagicMock(name="PropertyManager object")

    @mock.patch("restalchemy.dm.properties.PropertyManager", return_value=PM_MOCK)
    def setUp(self, pm_mock):
        super(ModelTestCase, self).setUp()
        self.PM_MOCK.__getitem__.side_effect = None
        self.PM_MOCK.reset_mock()
        self.pm_mock = pm_mock
        self.kwargs = {"kwarg1": 1, "kwarg2": 2}
        self.test_instance = models.Model(**self.kwargs)
        self.test_instance.validate = mock.MagicMock()

    def test_validate_call(self):
        models.Model.validate = mock.MagicMock()
        test_instance = models.Model(**self.kwargs)
        test_instance.validate.assert_called_once_with()

    def test_obj(self):
        self.assertEqual(self.test_instance.properties, self.PM_MOCK)
        self.pm_mock.assert_called_once_with(models.Model.properties, **self.kwargs)

    def test_obj_getattr(self):
        self.assertEqual(
            self.test_instance.fake_prop1, self.PM_MOCK["fake_prop1"].value
        )

    def test_obj_getattr_raise_attribute_error(self):
        self.PM_MOCK.__getitem__.side_effect = KeyError

        self.assertRaises(AttributeError, lambda: self.test_instance.fake_prop1)

    def test_obj_setattr(self):
        self.test_instance.fake_prop1 = FAKE_VALUE1

        self.assertEqual(self.PM_MOCK.__getitem__().value, FAKE_VALUE1)


class Model1(models.Model):
    pass


class Model2(models.Model):
    pass


class Model3(models.ModelWithUUID):
    pass


class BaseModel(models.Model):
    property1 = properties.property(types.Integer())
    property2 = properties.property(types.Integer())
    property3 = relationships.relationship(Model1)
    property4 = relationships.relationship(Model2)


class FakeModel(BaseModel):
    property1 = properties.property(types.String())
    property3 = relationships.relationship(Model3)


class SimpleViewModel(models.ModelWithUUID, models.SimpleViewMixin):
    int_property = properties.property(types.Integer(), default=1)
    str_property = properties.property(types.String(), default="foo")
    none_property = properties.property(types.AllowNone(types.Integer()), default=None)


class InheritModelTestCase(base.BaseTestCase):
    def test_correct_type_in_base_model(self):
        props = BaseModel.properties.properties

        self.assertIsInstance(props["property1"]._property_type, types.Integer)
        self.assertIsInstance(props["property2"]._property_type, types.Integer)
        self.assertEqual(props["property3"]._property_type, Model1)
        self.assertEqual(props["property4"]._property_type, Model2)

    def test_correct_type_in_inherit_model(self):
        props = FakeModel.properties.properties

        self.assertIsInstance(props["property1"]._property_type, types.String)
        self.assertIsInstance(props["property2"]._property_type, types.Integer)
        self.assertEqual(props["property3"]._property_type, Model3)
        self.assertEqual(props["property4"]._property_type, Model2)


class DirtyModelTestCase(base.BaseTestCase):
    def setUp(self):
        super(DirtyModelTestCase, self).setUp()
        self._model = FakeModel(
            property1="fake_string",
            property2=2,
            property3=Model3(),
            property4=Model2(),
        )

    def test_dirty_is_false(self):
        self.assertFalse(self._model.is_dirty())

    def test_dirty_is_false_after_change_property2(self):
        self._model.property2 = 6
        self._model.property2 = 2

        self.assertFalse(self._model.is_dirty())

    def test_property1_is_dirty(self):
        self._model.property1 = "new fake_string"

        self.assertTrue(self._model.is_dirty())

    def test_property3_is_dirty(self):
        self._model.property3 = Model3()

        self.assertTrue(self._model.is_dirty())

    def test_property3_is_not_dirty(self):
        self._model.property3 = Model3(uuid=self._model.property3.uuid)

        self.assertFalse(self._model.is_dirty())


class FakeModelWithID(BaseModel):
    uuid = properties.property(types.UUID(), id_property=True)
    property3 = relationships.relationship(Model3)


class FakeModelWithSeveralIDs(BaseModel):
    uuid = properties.property(types.UUID(), id_property=True)
    uuid2 = properties.property(types.UUID(), id_property=True)
    property3 = relationships.relationship(Model3)


class ModelWithIDsTestCase(base.BaseTestCase):
    def test_get_id_property(self):
        props = FakeModelWithID.properties.properties

        self.assertEqual(FakeModelWithID.get_id_property(), {"uuid": props["uuid"]})
        with self.assertRaises(TypeError):
            FakeModelWithSeveralIDs.get_id_property()

    def test_get_id_property_name(self):
        self.assertEqual(FakeModelWithID.get_id_property_name(), "uuid")
        with self.assertRaises(TypeError):
            FakeModelWithSeveralIDs.get_id_property_name()

    def test_model_plain_dict(self):
        fakeInt = 1
        fakeUUID1 = uuid.uuid4()
        fakeUUID2 = uuid.uuid4()
        fakeUUID3 = uuid.uuid4()

        model = FakeModelWithID(
            uuid=fakeUUID1,
            property1=fakeInt,
            property2=fakeInt,
            property3=Model3(uuid=fakeUUID2),
            property4=Model2(),
        )

        expected = {
            "uuid": fakeUUID1,
            "property1": fakeInt,
            "property2": fakeInt,
            "property3": fakeUUID2,
            "property4": None,
        }

        plain_dict = model.as_plain_dict()
        self.assertEqual(plain_dict, expected)

        # dictionary changes do not affect model
        plain_dict["property3"] = Model3(uuid=fakeUUID3)
        self.assertEqual(model.property3.uuid, fakeUUID2)


class ModelWithRequiredUUIDTestCase(base.BaseTestCase):
    def test_uuid_provided(self):
        model_uuid = uuid.uuid4()

        model = models.ModelWithRequiredUUID(uuid=model_uuid)

        self.assertEqual(model_uuid, model.uuid)

    def test_uuid_not_provided(self):
        self.assertRaises(exceptions.PropertyRequired, models.ModelWithRequiredUUID)

    def test_uuid_is_readonly(self):
        uuid_1 = uuid.uuid4()
        uuid_2 = uuid.uuid4()

        model = models.ModelWithRequiredUUID(uuid=uuid_1)

        with self.assertRaises(exceptions.ReadOnlyProperty):
            model.uuid = uuid_2

    def test_get_model_id(self):
        uuid_1 = uuid.uuid4()

        model = models.ModelWithRequiredUUID(uuid=uuid_1)

        self.assertEqual(model.get_id(), uuid_1)


class SimpleViewMixinTestCase(base.BaseTestCase):
    def test_dump_to_simple_view(self):
        simple_view_model = SimpleViewModel()
        view = simple_view_model.dump_to_simple_view()

        self.assertEqual(view["int_property"], 1)
        self.assertEqual(view["str_property"], "foo")
        self.assertIs(view["none_property"], None)
        self.assertIsInstance(view["uuid"], str)

    def test_restore_to_simple_view(self):
        simple_view_model = SimpleViewModel()
        view = {
            "uuid": "2e39d8df-2662-4834-ad86-c637d1edd504",
            "int_property": 2,
            "str_property": "bar",
            "none_property": None,
        }
        simple_view_model = SimpleViewModel.restore_from_simple_view(**view)

        self.assertEqual(simple_view_model.int_property, 2)
        self.assertEqual(simple_view_model.str_property, "bar")
        self.assertIs(simple_view_model.none_property, None)
        self.assertIs(simple_view_model.uuid.__class__, uuid.UUID)



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from restalchemy.storage.sql.dialect.query_builder import q
from restalchemy.storage.sql import filters as sql_filters


class AbstractProcessResult(metaclass=abc.ABCMeta):
    def __init__(self, result, session):
        self._result = result
        self._session = session

    def get_count(self):
        """
        Retrieves the number of rows affected by the query.

        :return: The number of rows affected by the executed SQL statement.
        :rtype: int
        """

        return self._result.rowcount

    @property
    def rows(self):
        """
        Retrieves all rows from the executed SQL statement.

        :return: All rows from the executed SQL statement.
        :rtype: list
        """
        return self.get_rows()

    @abc.abstractmethod
    def fetchall(self):
        """
        Retrieves all rows from the executed SQL statement.

        :return: All rows from the executed SQL statement.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_rows(self):
        """
        Retrieves all rows from the executed SQL statement.

        :return: All rows from the executed SQL statement.
        :rtype: list
        """
        raise NotImplementedError()


class BaseProcessResult(AbstractProcessResult):
    def __init__(self, result, session):
        """
        Initializes the BaseProcessResult object.

        :param result: The result of the executed SQL statement.
        :param session: The session object used for the execution
            of the SQL statement.
        """
        super(BaseProcessResult, self).__init__(result, session)
        self._rows = None

    def fetchall(self):
        """
        Retrieves all rows from the executed SQL statement.

        :return: All rows from the executed SQL statement.
        """
        for row in self._result:
            yield row

    def get_rows(self):
        """
        Retrieves all rows from the executed SQL statement and caches them.

        If the rows have not been previously retrieved, it fetches all rows
        from the result and stores them in the `_rows` attribute. Otherwise,
        it returns the cached rows.

        :return: All rows from the executed SQL statement.
        :rtype: list
        """

        if self._rows is None:
            self._rows = self._result.fetchall()
        return self._rows


class BaseOrmProcessResult(BaseProcessResult):
    def __init__(self, result, query, session):
        """
        Initializes the BaseOrmProcessResult object.

        :param result: The result of the executed ORM query
        :param query: The ORM query object used for the execution
            of the SQL statement
        :param session: The session object used for the execution
            of the SQL statement
        """
        super(BaseOrmProcessResult, self).__init__(
            result=result,
            session=session,
        )
        self._query = query
        self._rows = None

    def fetchall(self):
        """
        Retrieves all rows from the executed ORM query.

        :return: All rows from the executed ORM query, parsed into objects.
        """
        for row in self._result:
            yield self._query.parse_row(row)

    def get_rows(self):
        """
        Retrieves all rows from the executed ORM query, parsing them into
        objects and caching the result for future access.

        If the rows have not been previously retrieved, it fetches all rows
        from the result, parses them using the query's result parser, and
        stores them in the `_rows` attribute. Otherwise, it returns the cached
        rows.

        :return: All rows from the executed ORM query, parsed into objects.
        :rtype: list
        """

        if self._rows is None:
            self._rows = self._query.parse_results(self._result.fetchall())
        return self._rows


class AbstractDialectCommand(metaclass=abc.ABCMeta):
    def __init__(self, table, data, session):
        """
        Initializes the AbstractDialectCommand with the given table, data,
        and session.

        :param table: The table associated with the command.
        :param data: The data to be used in the command.
        :param session: The session to be used for executing the command.
        """

        self._table = table
        self._data = data
        self._session = session

    @abc.abstractmethod
    def get_values(self):
        """
        Retrieves the values to be used in the SQL command execution.

        This method should be implemented by subclasses to provide the logic
        for extracting or constructing the values required for the SQL
        statement. The values should be returned in a format compatible with
        the SQL execution context.

        :return: A collection of values to be used in the SQL command.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def get_statement(self):
        """
        Retrieves the SQL statement to be used in the command execution.

        This method should be implemented by subclasses to provide the logic
        for constructing the SQL statement required for the command. The
        statement should be returned as a string, with placeholders for the
        values to be inserted.

        :return: The SQL statement to be used in the command execution.
        :rtype: str
        """
        raise NotImplementedError()

    def execute(self):
        """
        Executes the SQL command using the values and statement provided by
        the abstract methods.

        :return: A ProcessResult object containing the result of the SQL
            command.
        """
        values = self.get_values()
        statement = self.get_statement()
        return BaseProcessResult(
            result=self._session.execute(statement, values),
            session=self._session,
        )


class BaseInsertCommand(AbstractDialectCommand):
    EXPRESSION = "INSERT INTO `%s` (%s) VALUES (%s)"

    def get_values(self):
        """
        Retrieves the values to be inserted into the SQL command.

        This method iterates over the column names of the table and collects
        the corresponding data values into a tuple. These values are then
        used as the data to be inserted into the SQL statement.

        :return: A tuple of values corresponding to the column names of the
            table.
        :rtype: tuple
        """

        values = tuple()
        for column_name in self._table.get_column_names(self._session):
            values += (self._data[column_name],)
        return values

    def get_statement(self):
        """
        Retrieves the SQL statement to be used in the command execution.

        This method constructs the SQL statement to be used in the command
        execution by inserting the table name and column names into the
        EXPRESSION template string. The column names are escaped according to
        the dialect's escaping rules, and the placeholders for the values are
        created by repeating the string "%s" for each column.

        :return: The SQL statement to be used in the command execution.
        :rtype: str
        """
        column_names = self._table.get_escaped_column_names(self._session)
        return self.EXPRESSION % (
            self._table.name,
            ", ".join(column_names),
            ", ".join(["%s"] * len(column_names)),
        )


class BaseUpdateCommand(AbstractDialectCommand):
    EXPRESSION = "UPDATE `%s` SET %s WHERE %s"

    def __init__(self, table, ids, data, session):
        """
        Initializes the BaseUpdateCommand with the given table, ids, data,
        and session.

        :param table: The table associated with the command.
        :param ids: The ids to be updated.
        :param data: The data to be used in the command.
        :param session: The session to be used for executing the command.
        """
        super().__init__(table, data, session=session)
        self._ids = ids

    def get_values(self):
        """
        Retrieves the values to be updated in the SQL command.

        This method iterates over the column names of the table (excluding the
        primary key) and collects the corresponding data values into a tuple.
        Additionally, it collects the primary key values from the `_ids` attr
        into the same tuple. These values are then used as the data to be
        updated in the SQL statement.

        :return: A tuple of values corresponding to the column names of the
            table.
        :rtype: tuple
        """
        values = tuple()
        column_names = self._table.get_column_names(
            self._session,
            with_pk=False,
        )
        pk_names = self._table.get_pk_names(session=self._session)
        for column_name in column_names:
            values += (self._data[column_name],)
        for column_name in pk_names:
            values += (self._ids[column_name],)
        return values

    def get_statement(self):
        """
        Retrieves the SQL statement to be used in the command execution.

        This method constructs the SQL statement to be used in the command
        execution by inserting the table name, column names, and primary key
        names into the EXPRESSION template string. The column names are escaped
        according to the dialect's escaping rules, and the placeholders for the
        values are created by repeating the string "%s" for each column.

        :return: The SQL statement to be used in the command execution.
        :rtype: str
        """
        column_names = self._table.get_escaped_column_names(
            session=self._session,
            with_pk=False,
        )
        pk_names = self._table.get_escaped_pk_names(session=self._session)
        return self.EXPRESSION % (
            self._table.name,
            ", ".join([f"{name} = %s" for name in column_names]),
            " AND ".join([f"{name} = %s" for name in pk_names]),
        )


class BaseDeleteCommand(AbstractDialectCommand):
    EXPRESSION = "DELETE FROM `%s` WHERE %s"

    def __init__(self, table, ids, session):
        """
        Initializes the BaseDeleteCommand with the given table, ids, and
        session.

        :param table: The table associated with the command.
        :param ids: The ids to be used for deleting the rows.
        :param session: The session to be used for executing the command.
        """
        super().__init__(table=table, data={}, session=session)
        self._ids = ids

    def get_values(self):
        """
        Retrieves the values to be used in the SQL command execution.

        This method iterates over the primary key column names of the table
        and collects the corresponding data values into a tuple. These
        values are then used as the data to be deleted in the SQL statement.

        :return: A tuple of values corresponding to the primary key column
            names of the table.
        :rtype: tuple
        """
        values = tuple()
        pk_names = self._table.get_pk_names(session=self._session)
        for column_name in pk_names:
            values += (self._ids[column_name],)
        return values

    def get_statement(self):
        """
        Retrieves the SQL statement to be used in the command execution.

        This method constructs the SQL statement to be used in the command
        execution by inserting the table name and primary key column names
        into the EXPRESSION template string.

        :return: The SQL statement to be used in the command execution.
        :rtype: str
        """
        pk_names = self._table.get_escaped_pk_names(session=self._session)
        return self.EXPRESSION % (
            self._table.name,
            " AND ".join([f"{name} = %s" for name in pk_names]),
        )


class BaseBatchDelete(AbstractDialectCommand):
    EXPRESSION_IN = "DELETE FROM `%s` WHERE %s in %s"
    EXPRESSION_FILTER = "DELETE FROM `%s` WHERE %s"

    def __init__(self, table, snapshot, session):
        """
        Initializes the BaseBatchDelete with the given table, snapshot, and
        session.

        :param table: The table associated with the command.
        :param snapshot: The snapshot containing the rows to be deleted.
        :param session: The session to be used for executing the command.
        """
        super().__init__(
            table=table,
            data={},
            session=session,
        )
        self._snapshot = snapshot
        self._pk_keys = self._table.get_escaped_pk_names(session=session)
        keys_count = len(self._pk_keys)
        if keys_count == 1:
            self._is_multiple_primary_key = False
        elif keys_count > 1:
            self._is_multiple_primary_key = True
        else:
            raise ValueError(f"The model with table {table!r} has 0 primary keys")

    def _get_values(self):
        """
        Retrieves the values to be used in the SQL command execution.

        This method iterates over the snapshot rows and collects the values
        corresponding to the primary key column names of the table into a
        list. These values are then used as the data to be deleted in the SQL
        statement.

        :return: A list of values corresponding to the primary key column
            names of the table.
        :rtype: list
        """
        values = []
        for snapshot in self._snapshot:
            for key in self._table.get_pk_names(session=self._session):
                values.append(snapshot[key])
        return values

    def _get_multiple_primary_key_values(self):
        """
        Retrieves the values to be used in the SQL command execution when
        multiple primary keys are declared for the table.

        :return: A list of values corresponding to the primary key column
            names of the table.
        :rtype: list
        """
        return self._get_values()

    def _get_single_primary_key_values(self):
        """
        Retrieves the primary key values for tables with a single primary key.

        This method wraps the result of `_get_multiple_primary_key_values` in a
        list to optimize the `in` operation for SQL command execution.

        :return: A list containing the primary key values.
        :rtype: list
        """
        # NOTE(efrolov): Wrap to list for `in` optimization
        return [self._get_multiple_primary_key_values()]

    def get_values(self):
        """
        Retrieves the values to be used in the SQL command execution.

        Depending on the presence of multiple primary keys declared for the
        table,this method returns either a list of values corresponding to
        the primary key column names of the table or a list containing the
        primary key values for tables with a single primary key.

        :return: List of values to be used in the SQL command execution.
        :rtype: list
        """
        return (
            self._get_multiple_primary_key_values()
            if self._is_multiple_primary_key
            else self._get_single_primary_key_values()
        )

    def _get_single_primary_key_statement(self):
        """
        Retrieves the SQL statement to be used in the command execution
        when a single primary key is declared for the table.

        This method constructs the SQL statement to be used in the command
        execution by inserting the table name and primary key column name
        into the EXPRESSION_IN template string.

        :return: The SQL statement to be used in the command execution.
        :rtype: str
        """
        return self.EXPRESSION_IN % (
            self._table.name,
            self._pk_keys[0],
            "%s",
        )

    def _get_multiple_primary_key_statement(self):
        """
        Retrieves the SQL statement to be used in the command execution
        when multiple primary keys are declared for the table.

        This method constructs the SQL statement to be used in the command
        execution by inserting the table name and primary key column names
        into the EXPRESSION_FILTER template string. The `where_condition` is
        constructed by joining the condition for each primary key column
        with the `OR` operator.

        :return: The SQL statement to be used in the command execution.
        :rtype: str
        """
        where_part = " AND ".join([f"{key} = %s" for key in self._pk_keys])
        where_condition = " OR ".join([where_part for _ in range(len(self._snapshot))])
        return self.EXPRESSION_FILTER % (
            self._table.name,
            where_condition,
        )

    def get_statement(self):
        """
        Retrieves the SQL statement to be used in the command execution.

        This method constructs the SQL statement based on whether the table
        has multiple primary keys or a single primary key. It selects the
        appropriate expression template and formats it with the table name
        and primary key conditions.

        :return: The SQL statement to be used in the command execution.
        :rtype: str
        """

        return (
            self._get_multiple_primary_key_statement()
            if self._is_multiple_primary_key
            else self._get_single_primary_key_statement()
        )


class BaseBasicSelectCommand(AbstractDialectCommand):
    EXPRESSION = "SELECT %s FROM `%s`"

    def __init__(
        self,
        table,
        session,
        limit=None,
        order_by=None,
        locked=False,
    ):
        """
        Initializes the BaseBasicSelectCommand with the given parameters.

        :param table: The table from which data is to be selected.
        :param session: The session to be used for executing the command.
        :param limit: The maximum number of rows to return. Optional.
        :type limit: int, optional
        :param order_by: A dictionary specifying the columns to order by and
            their sort types. Optional.
        :param locked: Whether to lock the selected rows for update. Defaults
            to False.
        :type locked: bool
        """

        super().__init__(table=table, data={}, session=session)
        self._limit = limit
        self._order_by = order_by
        self._locked = locked

    def construct_limit(self):
        """
        Constructs the LIMIT clause for the SQL statement based on the given
        limit.

        :return: The LIMIT clause for the SQL statement.
        :rtype: str
        """
        if self._limit:
            return " LIMIT " + str(self._limit)
        return ""

    def construct_locked(self):
        """
        Constructs the FOR UPDATE clause for the SQL statement based on the
        given locked status.

        :return: The FOR UPDATE clause for the SQL statement.
        :rtype: str
        """
        if self._locked:
            return " FOR UPDATE"
        return ""

    def construct_order_by(self):
        """
        Constructs the ORDER BY clause for the SQL statement based on the
        given order_by.

        :return: The ORDER BY clause for the SQL statement.
        :rtype: str
        """

        if self._order_by:
            res = []
            for name, sorttype in self._order_by.items():
                sorttype = sorttype.upper()
                if sorttype not in ["ASC", "DESC", "", None]:
                    raise ValueError(f"Unknown order: {sorttype}.")
                res.append(f"{self._session.engine.escape(name)} {sorttype or 'ASC'}")
            return " ORDER BY " + ", ".join(res)
        return ""


class BaseSelectCommand(BaseBasicSelectCommand):
    def __init__(
        self,
        table,
        session,
        filters=None,
        limit=None,
        order_by=None,
        locked=False,
    ):
        """
        Initializes the BaseSelectCommand with the given parameters.

        :param table: The table from which data is to be selected.
        :param session: The session to be used for executing the command.
        :param filters: The filters to be applied to the query. Optional.
        :param limit: The maximum number of rows to return. Optional.
        :type limit: int, optional
        :param order_by: A dictionary specifying the columns to order by
            and their sort types. Optional.
        :param locked: Whether to lock the selected rows for update. Defaults
            to False.
        :type locked: bool
        """
        super(BaseSelectCommand, self).__init__(
            table=table,
            session=session,
            limit=limit,
            order_by=order_by,
            locked=locked,
        )
        self._filters = sql_filters.convert_filters(
            model=self._table.model,
            filters_root=filters,
            session=session,
        )

    def get_values(self):
        """
        Retrieves the values to be used in the SQL command execution.

        This method returns the values part of the WHERE clause of the
        SQL statement, which is used to filter the results of the query.

        :return: The values part of the WHERE clause.
        :rtype: tuple
        """

        return self._filters.value

    def construct_where(self):
        """
        Constructs the WHERE clause part of the SQL statement.

        This method returns the WHERE clause part of the SQL statement,
        which is used to filter the results of the query.

        :return: The WHERE clause part of the SQL statement.
        :rtype: str
        """
        return self._filters.construct_expression()

    def get_statement(self):
        sql = self.EXPRESSION % (
            ", ".join(self._table.get_escaped_column_names(self._session)),
            self._table.name,
        )
        filt = self.construct_where()

        return (
            (f"{sql} WHERE {filt}" if filt else sql)
            + self.construct_order_by()
            + self.construct_limit()
            + self.construct_locked()
        )


class BaseCustomSelectCommand(BaseBasicSelectCommand):
    def __init__(
        self,
        table,
        where_conditions,
        where_values,
        session,
        limit=None,
        order_by=None,
        locked=False,
    ):
        """
        :param table: Table to select from
        :type table: :class:`~restalchemy.storage.sql.tables.SQLTable`
        :param where_conditions: Conditions for the WHERE clause in the SQL
            statement.
        :param where_values: Values for the WHERE clause in the SQL statement.
        :type where_values: tuple
        :param session: Session to use for the query
        :type session: :class:`~restalchemy.storage.sql.sessions.SQLSession`
        :param limit: Maximum number of records to return
        :type limit: int
        :param order_by: Fields to order the records by
        :type order_by: dict
        :param locked: Whether to lock the records using a FOR UPDATE clause
        :type locked: bool
        """
        super().__init__(
            table=table,
            session=session,
            limit=limit,
            order_by=order_by,
            locked=locked,
        )
        self._where_conditions = where_conditions
        self._where_values = where_values

    def get_values(self):
        """
        Retrieves the values to be used in the SQL command execution.

        This method returns the values part of the WHERE clause of the
        SQL statement, which is used to filter the results of the query.

        :return: The values part of the WHERE clause.
        :rtype: tuple
        """
        return self._where_values

    def construct_where(self):
        """
        Constructs the WHERE clause part of the SQL statement.

        This method returns the WHERE clause part of the SQL statement,
        which is used to filter the results of the query.

        :return: The WHERE clause part of the SQL statement.
        :rtype: str
        """
        return self._where_conditions

    def get_statement(self):
        """
        Retrieves the full SQL statement for the query.

        :return: The full SQL statement for the query.
        :rtype: str
        """
        sql = self.EXPRESSION % (
            ", ".join(
                self._table.get_escaped_column_names(
                    self._session,
                )
            ),
            self._table.name,
        )
        return (
            sql
            + " WHERE "
            + self.construct_where()
            + self.construct_order_by()
            + self.construct_limit()
            + self.construct_locked()
        )


class BaseCountCommand(BaseSelectCommand):
    EXPRESSION = "SELECT COUNT(*) as count FROM `%s`"

    def __init__(self, table, session, filters=None):
        """
        Initializes the BaseCountCommand with the provided table, session,
        and optional filters.

        :param table: The table from which to perform the count query.
        :param session: The session used to execute the SQL command.
        :param filters: Optional filters to apply to the count query.
        :type filters: Any, optional
        """

        super().__init__(table=table, session=session, filters=filters)

    def get_statement(self):
        """
        Retrieves the full SQL statement for the query.

        :return: The full SQL statement for the query.
        :rtype: str
        """
        sql = self.EXPRESSION % (self._table.name)
        filt = self.construct_where()

        return f"{sql} WHERE {filt}" if filt else sql


class BaseOrmDialectCommand(AbstractDialectCommand):
    def __init__(self, table, query, session):
        """
        Initializes the BaseOrmDialectCommand with the given table, query, and
        session.

        :param table: The table to be used in the ORM dialect command.
        :param query: The query object that contains the SQL query to be
            compiled.
        :param session: The session to be used for executing the command.
        """

        super().__init__(table=table, session=session, data=None)
        self._query = query

    def get_statement(self):
        """
        Compiles and retrieves the SQL statement for the query.

        This method utilizes the `compile` method of the query object
        to generate the SQL statement that represents the query.

        :return: The compiled SQL statement for the query.
        :rtype: str
        """

        return self._query.compile()

    def get_values(self):
        """
        Retrieves the values to be used in the SQL command execution.

        This method should be implemented by subclasses to provide the logic
        for extracting or constructing the values required for the SQL
        statement. The values should be returned in a format compatible with
        the SQL execution context.

        :return: A collection of values to be used in the SQL command.
        :rtype: iterable
        """
        return self._query.values()

    def execute(self):
        """
        Executes the SQL command associated with the query.

        This method compiles the query using the `get_statement` method and
        retrieves the values required for the query execution using the
        `get_values` method. Subsequently, it executes the SQL command using
        the parent class's `execute` method, passing the compiled statement and
        values to it. The result of the execution is then wrapped in a
        `BaseOrmProcessResult` object and returned.

        :return: The result of the query execution, wrapped in a
            `BaseOrmProcessResult` object.
        :rtype: BaseOrmProcessResult
        """
        return BaseOrmProcessResult(
            result=super().execute(),
            query=self._query,
            session=self._session,
        )


class BaseSqlOrm(object):
    @staticmethod
    def select(model, session):
        """
        Creates a new Q object for the given model and session.

        :param model: The model class for which the Q object is to be created.
        :param session: The session to be used for executing the query.
        :return: A new Q object instance.
        """
        return q.Q.select(model, session)


class AbstractDialect(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def DIALECT_NAME(self):
        """
        The name of the dialect as a string.

        This property must be overridden by subclasses to provide a valid
        dialect name.

        :return: The name of the dialect as a string.
        :rtype: str
        """
        raise NotImplementedError()

    def __init__(self):
        """
        Initializes the AbstractDialect instance.

        This constructor sets up the base SQL ORM for the dialect by
        creating an instance of BaseSqlOrm.

        """

        super(AbstractDialect, self).__init__()
        self._orm = BaseSqlOrm()

    @property
    def orm(self):
        """
        Retrieves the BaseSqlOrm instance associated with the dialect.

        This property returns the BaseSqlOrm instance that is used by the
        dialect for ORM operations.

        :return: The BaseSqlOrm instance associated with the dialect.
        :rtype: BaseSqlOrm
        """
        return self._orm

    @property
    def name(self):
        """
        Retrieves the name of the dialect as a string.

        This property returns the string value of the DIALECT_NAME property of
        the dialect.

        :return: The name of the dialect as a string.
        :rtype: str
        """
        return self.DIALECT_NAME

    @abc.abstractmethod
    def insert(self, table, data, session):
        """
        Inserts data into the specified table within the session context.

        This method should be implemented by subclasses to provide the logic
        for inserting the provided data into the given table using the
        specified session. The method should construct and execute an SQL
        insert command appropriate for the dialect.

        :param table: The table into which data is to be inserted.
        :param data: The data to be inserted into the table.
        :param session: The session to be used for executing the insert
            command.
        :raises NotImplementedError: If the method is not implemented by a
            subclass.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def update(self, table, ids, data, session):
        """
        Updates existing records in the specified table with the provided data
        within the session context.

        This method should be implemented by subclasses to provide the logic
        for updating existing records in the given table using the specified
        session. The method should construct and execute an SQL update command
        appropriate for the dialect.

        :param table: The table to be updated.
        :param ids: The IDs of the records to be updated.
        :type ids: list
        :param data: The data to be used for updating the records.
        :param session: The session to be used for executing the update
            command.
        :raises NotImplementedError: If the method is not implemented by a
            subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def delete(self, table, ids, session):
        """
        Deletes records from the specified table based on the provided IDs
        within the session context.

        This method should be implemented by subclasses to provide the logic
        for deleting records from the given table using the specified session.
        The method should construct and execute an SQL delete command
        appropriate for the dialect.

        :param table: The table from which records are to be deleted.
        :param ids: The IDs of the records to be deleted.
        :param session: The session to be used for executing the delete
            command.
        :raises NotImplementedError: If the method is not implemented by a
            subclass.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def select(
        self,
        table,
        filters,
        session,
        limit=None,
        order_by=None,
        locked=False,
    ):
        """
        Retrieves records from the specified table based on the provided
        filters and ordering criteria, with optional limit and locking
        constraints, within the session context.

        This method should be implemented by subclasses to provide the logic
        for selecting records from the given table using the specified
        session. The method should construct and execute an SQL select command
        appropriate for the dialect.

        :param table: The table from which records are to be retrieved.
        :param filters: The filters to be used for selecting the records.
        :param session: The session to be used for executing the select
            command.
        :param limit: The maximum number of records to be retrieved.
        :type limit: int
        :param order_by: The ordering criteria to be used for selecting the
            records.
        :param locked: Whether or not to lock the records for this session.
        :type locked: bool
        :raises NotImplementedError: If the method is not implemented by a
            subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def count(self, table, filters, session):
        """
        Retrieves the number of records in the specified table that match the
        given filters within the session context.

        This method should be implemented by subclasses to provide the logic
        for counting the records from the given table using the specified
        session. The method should construct and execute an SQL count command
        appropriate for the dialect.

        :param table: The table from which records are to be counted.
        :param filters: The filters to be used for counting the records.
        :param session: The session to be used for executing the count command.
        :raises NotImplementedError: If the method is not implemented by a
            subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def orm_command(self, table, query, session):
        """
        Creates a new ORM command for the given table, query, and session.

        This method should be implemented by subclasses to provide the logic
        for creating an ORM command that is appropriate for the dialect.

        :param table: The table to be used in the ORM dialect command.
        :param query: The query object that contains the SQL query to be
            compiled.
        :param session: The session to be used for executing the command.
        :raises NotImplementedError: If the method is not implemented by a
            subclass.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def custom_select(
        self,
        table,
        session,
        where_conditions,
        where_values,
        limit=None,
        order_by=None,
        locked=False,
    ):
        """
        Constructs and executes a custom SQL select command for the specified
        table, using the given session and where conditions.

        This method should be implemented by subclasses to provide the logic
        for selecting records based on custom where conditions and values, with
        optional limit, order by, and locking constraints.

        :param table: The table from which records are to be selected.
        :param session: The session to be used for executing the select
            command.
        :param where_conditions: The conditions for the WHERE clause in the SQL
            statement.
        :param where_values: The values for the WHERE clause in the SQL
            statement.
        :param limit: The maximum number of records to be retrieved.
        :type limit: int, optional
        :param order_by: The ordering criteria for the selected records.
        :param locked: Whether or not to lock the selected records for update.
        :type locked: bool
        :raises NotImplementedError: If the method is not implemented by a
            subclass.
        """

        raise NotImplementedError()



#    Copyright 2021 Eugene Frolov.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.common import contexts
from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.storage.sql import engines
from restalchemy.storage.sql import orm
from restalchemy.storage.sql import sessions


# CREATE TABLE `foos` (
#      `uuid` CHAR(36) CHARACTER SET utf8 COLLATE utf8_unicode_ci NOT NULL ,
#      `foo_field1` INT NOT NULL ,
#      `foo_field2` VARCHAR(255) NOT NULL ,
# PRIMARY KEY (`uuid`(36)) USING HASH)
# ENGINE = InnoDB;
class FooModel(models.ModelWithUUID, orm.SQLStorableMixin):
    __tablename__ = "foos"
    foo_field1 = properties.property(types.Integer(), required=True)
    foo_field2 = properties.property(types.String(), default="foo_str")


# One must set name 'default' for default storage if you want to work with
# storage operations without session parameter.
engines.engine_factory.configure_factory(
    db_url="mysql://user:password@127.0.0.1:3306/test",
    name="default",  # name argument by default is 'default' and can be skipped
)

# db_one engine is not equal default engine because it has separate connection
# pool therefore session from default engine is not equal session from
# db_one engine. But if you use thread storage sessions any storage operations
# will working with session from thread storage by default.
engines.engine_factory.configure_factory(
    db_url="mysql://user:password@127.0.0.1:3306/test",
    name="db_one",  # by default name is 'default'
)
engines.engine_factory.configure_factory(
    db_url="mysql://user:password@127.0.0.1:3307/test",
    name="db_two",  # by default name is 'default'
)

engine_one = engines.engine_factory.get_engine("db_one")
engine_two = engines.engine_factory.get_engine("db_two")

session_one = engine_one.get_session()
session_two = engine_two.get_session()

foo_one = FooModel(foo_field1=1)
foo_one.insert(session=session_one)
session_one.commit()


foo_two = FooModel(foo_field1=2)
foo_two.insert(session=session_two)
session_two.commit()

with engine_two.session_manager() as session:
    # Incorrect. Session from context manager is not used because new session
    # from default storage going to create in get_all call. And you get data
    # from default database (engine).
    #
    # the result: [<FooModel {foo_field1: 1,
    #                         uuid: 27ce96dc-daad-4272-88d8-8e41028c38e9,
    #                         foo_field2: foo_str}>]
    print("Incorrect input for storage two:")
    print(FooModel.objects.get_all())

    # correct call
    # the result: [<FooModel {foo_field1: 2,
    #                         uuid: 78141610-6709-4296-bd56-7c62aec97aa8,
    #                         foo_field2: foo_str} >]
    print("Correct input for storage two:")
    print(FooModel.objects.get_all(session=session))


# The next expressions is correct
context_two = contexts.Context(engine_name="db_two")
with context_two.session_manager() as session:
    print("Correct input for storage two for context session:")
    print(FooModel.objects.get_all())


# But you can not create two active session from context
context_one = contexts.Context(engine_name="db_one")
with context_two.session_manager() as session_2:
    try:
        with context_one.session_manager() as session_1:
            print("Newer call")
    except sessions.SessionConflict as e:
        print("Session conflict:", e)



# Copyright 2014 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest


class BaseTestCase(unittest.TestCase):
    FAKE_STRING = "FakeString"

    def string_generator(self, size=6):
        return (self.FAKE_STRING * (size / len(self.FAKE_STRING) + 1))[:size]



# Copyright 2016 Eugene Frolov <eugene@frolov.net.ru>
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from restalchemy.storage.sql import migrations


class MigrationStep(migrations.AbstractMigrationStep):
    def __init__(self):
        self._depends = [""]

    @property
    def migration_id(self):
        return "f3841e0e-e180-4144-bd82-6113d646c2c9"

    def upgrade(self, session):
        expressions = [
            """
                CREATE TABLE IF NOT EXISTS root (
                    uuid CHAR(36) NOT NULL,
                    field_str VARCHAR(255) NOT NULL,
                    field_int INT NOT NULL,
                    field_bool BOOL NOT NULL,
                    PRIMARY KEY (uuid)
                );
            """,
            """
                CREATE TABLE IF NOT EXISTS lnp1_1 (
                    uuid CHAR(36) NOT NULL,
                    root CHAR(36) NOT NULL,
                    field_str VARCHAR(255) NOT NULL,
                    field_int INT NOT NULL,
                    field_bool BOOL NOT NULL,
                    PRIMARY KEY (uuid),
                    FOREIGN KEY (root)
                        REFERENCES root (uuid)
                        ON DELETE RESTRICT
                        ON UPDATE RESTRICT
                );
            """,
            """
                CREATE TABLE IF NOT EXISTS lnp1_2 (
                    uuid CHAR(36) NOT NULL,
                    root CHAR(36),
                    field_str VARCHAR(255) NOT NULL,
                    field_int INT NOT NULL,
                    field_bool BOOL NOT NULL,
                    PRIMARY KEY (uuid),
                    FOREIGN KEY (root)
                        REFERENCES root (uuid)
                        ON DELETE RESTRICT
                        ON UPDATE RESTRICT
                );
            """,
            """
                CREATE TABLE IF NOT EXISTS lwp1_1 (
                    uuid CHAR(36) NOT NULL,
                    root CHAR(36) NOT NULL,
                    field_str VARCHAR(255) NOT NULL,
                    field_int INT NOT NULL,
                    field_bool BOOL NOT NULL,
                    PRIMARY KEY (uuid),
                    FOREIGN KEY (root)
                        REFERENCES root (uuid)
                        ON DELETE RESTRICT
                        ON UPDATE RESTRICT
                );
            """,
            """
                CREATE TABLE IF NOT EXISTS lwp1_2 (
                    uuid CHAR(36) NOT NULL,
                    root CHAR(36),
                    field_str VARCHAR(255) NOT NULL,
                    field_int INT NOT NULL,
                    field_bool BOOL NOT NULL,
                    PRIMARY KEY (uuid),
                    FOREIGN KEY (root)
                        REFERENCES root (uuid)
                        ON DELETE RESTRICT
                        ON UPDATE RESTRICT
                );
            """,
            """
                CREATE TABLE IF NOT EXISTS lwp2_1 (
                    uuid CHAR(36) NOT NULL,
                    lwp1_1 CHAR(36) NOT NULL,
                    field_str VARCHAR(255) NOT NULL,
                    field_int INT NOT NULL,
                    field_bool BOOL NOT NULL,
                    PRIMARY KEY (uuid),
                    FOREIGN KEY (lwp1_1)
                        REFERENCES lwp1_1 (uuid)
                        ON DELETE RESTRICT
                        ON UPDATE RESTRICT
                );
            """,
            """
                CREATE TABLE IF NOT EXISTS lwp2_2 (
                    uuid CHAR(36) NOT NULL,
                    lwp1_1 CHAR(36),
                    field_str VARCHAR(255) NOT NULL,
                    field_int INT NOT NULL,
                    field_bool BOOL NOT NULL,
                    PRIMARY KEY (uuid),
                    FOREIGN KEY (lwp1_1)
                        REFERENCES lwp1_1 (uuid)
                        ON DELETE RESTRICT
                        ON UPDATE RESTRICT
                );
            """,
            """
                CREATE TABLE IF NOT EXISTS lwp2_3 (
                    uuid CHAR(36) NOT NULL,
                    lwp1_2 CHAR(36) NOT NULL,
                    field_str VARCHAR(255) NOT NULL,
                    field_int INT NOT NULL,
                    field_bool BOOL NOT NULL,
                    PRIMARY KEY (uuid),
                    FOREIGN KEY (lwp1_2)
                        REFERENCES lwp1_2 (uuid)
                        ON DELETE RESTRICT
                        ON UPDATE RESTRICT
                );
            """,
            """
                CREATE TABLE IF NOT EXISTS lwp2_4 (
                    uuid CHAR(36) NOT NULL,
                    lwp1_2 CHAR(36),
                    field_str VARCHAR(255) NOT NULL,
                    field_int INT NOT NULL,
                    field_bool BOOL NOT NULL,
                    PRIMARY KEY (uuid),
                    FOREIGN KEY (lwp1_2)
                        REFERENCES lwp1_2 (uuid)
                        ON DELETE RESTRICT
                        ON UPDATE RESTRICT
                );
            """,
        ]

        for expression in expressions:
            session.execute(expression)

    def downgrade(self, session):
        tables = [
            "lwp2_4",
            "lwp2_3",
            "lwp2_2",
            "lwp2_1",
            "lwp1_2",
            "lwp1_1",
            "lnp1_2",
            "lnp1_1",
            "root",
        ]

        for table in tables:
            self._delete_table_if_exists(session, table)


migration_step = MigrationStep()
